from collections import OrderedDict
from datetime import datetime, timedelta
from typing import (
    Optional,
    Any,
    Dict,
    List,
    Callable,
    Tuple,
    Literal,
    cast,
    Union,
    Iterable,
)

from phidata.base import PhidataBase, PhidataBaseArgs
from phidata.asset import DataAsset
from phidata.check import Check
from phidata.exceptions.task import TaskFailure
from phidata.task import Task, PythonTask, PythonTaskType
from phidata.task.task_relatives import TaskRelatives
from phidata.utils.context import get_run_date, build_path_context_from_env
from phidata.utils.cli_console import print_subheading, print_info
from phidata.utils.env_var import validate_env_vars
from phidata.utils.log import logger
from phidata.types.run_status import RunStatus
from phidata.types.context import (
    PathContext,
    RunContext,
    AirflowContext,
    DockerContext,
    K8sContext,
)


class WorkflowArgs(PhidataBaseArgs):
    # Tasks in this Workflow
    # Accepts a list of tasks or a list of lists of tasks
    tasks: Optional[List[Union[Task, List[Task]]]] = None
    # A dependency graph of the tasks in the format:
    # { downstream: [upstream_list] }
    # The downstream task will execute after all tasks in the
    #   upstream_list are finished.
    # upstream_list: tasks which run before the downstream task
    # Ex: To run extract -> transform -> load
    #   graph = {
    #     transform: [extract],
    #     load: [transform],
    #     ...
    #   }
    graph: Optional[Dict[Task, List[Task]]] = None
    # NOTE: Only for local execution
    # What to do when a task fails while running the workflow locally.
    # Continue to next task or stop?
    on_failure: Literal["stop", "continue"] = "stop"

    # Context
    # run_context provided by wf_operator
    run_context: Optional[RunContext] = None
    # path_context provided by env variables
    path_context: Optional[PathContext] = None
    # airflow_context provided by airflow
    airflow_context: Optional[AirflowContext] = None
    # Env vars that validate Airflow is active on the containers
    # This is used to gate code blocks which should only run on
    # airflow containers like creating DAGs and tasks
    validate_airflow_env: Dict[str, Any] = {"INIT_AIRFLOW": True}

    # Input DataAssets used to build sensors before kicking off the workflow
    inputs: Optional[List[DataAsset]] = None
    # DataAssets produced by this workflow, used for building the lineage graph
    outputs: Optional[List[DataAsset]] = None

    # Checks to run before the workflow
    pre_checks: Optional[List[Check]] = None
    # Checks to run after the workflow
    post_checks: Optional[List[Check]] = None

    # Add an EmptyOperator at the start of the workflow
    add_start_task: bool = False
    # Add an EmptyOperator at the end of the workflow
    add_end_task: bool = False

    # Airflow args
    # use airflow task_groups
    use_task_groups: bool = False
    # airflow task group_id for this workflow, use name if not provided
    group_id: Optional[str] = None
    # airflow dag_id for this workflow, use name if not provided
    dag_id: Optional[str] = None
    # The tooltip of the TaskGroup node when displayed in the UI
    description: Optional[str] = None
    # The fill color of the TaskGroup node when displayed in the UI
    ui_color: Optional[str] = None
    # The label color of the TaskGroup node when displayed in the UI
    ui_fgcolor: Optional[str] = None

    @property
    def airflow_active(self) -> bool:
        return validate_env_vars(self.validate_airflow_env)

    @property
    def run_date(self) -> datetime:
        return get_run_date(self.run_context, self.airflow_context)

    @classmethod
    def from_kwargs(cls, kwargs: Optional[Dict] = None):
        if kwargs is None or not isinstance(kwargs, dict):
            return cls()
        # logger.info(f"Loading {cls.__name__} using kwargs")
        args_object = cls(**kwargs)
        validate_airflow_env = kwargs.get("validate_airflow_env", None)
        if validate_env_vars(validate_airflow_env):
            # logger.info("Creating airflow_context")
            airflow_context = AirflowContext(**kwargs)
            args_object.airflow_context = airflow_context
        return args_object


class Workflow(PhidataBase):
    """Base Class for all Workflows"""

    def __init__(
        self,
        name: Optional[str] = None,
        # List of tasks in this Workflow
        # Tasks in this Workflow
        # Accepts a list of tasks or a list of lists of tasks
        tasks: Optional[List[Union[Task, List[Task]]]] = None,
        # A dependency graph of the tasks in the format:
        # { downstream: [upstream_list] }
        # The downstream task will execute only after all tasks in the
        #   upstream_list are finished.
        # upstream_list: tasks which run before the downstream task
        # Ex: To run extract -> transform -> load
        #   graph = {
        #     transform: [extract],
        #     load: [transform],
        #     ...
        #   }
        graph: Optional[Dict[Task, List[Task]]] = None,
        # NOTE: Only for local execution
        # What to do when a task fails while running the workflow locally.
        # Continue to next task or stop?
        on_failure: Literal["stop", "continue"] = "stop",
        # Context
        # run_context provided by wf_operator
        run_context: Optional[RunContext] = None,
        # path_context provided by env variables
        path_context: Optional[PathContext] = None,
        # airflow_context provided by airflow
        airflow_context: Optional[AirflowContext] = None,
        # Input DataAssets used to build sensors before kicking off the workflow
        inputs: Optional[List[DataAsset]] = None,
        # DataAssets produced by this workflow, used for building the lineage graph
        outputs: Optional[List[DataAsset]] = None,
        # Checks to run before the workflow
        pre_checks: Optional[List[Check]] = None,
        # Checks to run after the workflow
        post_checks: Optional[List[Check]] = None,
        # Add an EmptyOperator at the start of the workflow
        add_start_task: bool = False,
        # Add an EmptyOperator at the end of the workflow
        add_end_task: bool = False,
        # use airflow task_groups
        use_task_groups: bool = False,
        # Airflow args
        # airflow task group_id for this workflow, use name if not provided
        group_id: Optional[str] = None,
        # airflow dag_id for this workflow, use name if not provided
        dag_id: Optional[str] = None,
        # # The tooltip of the TaskGroup node when displayed in the UI
        # description: Optional[str] = None,
        # # The fill color of the TaskGroup node when displayed in the UI
        # ui_color: Optional[str] = None,
        # # The label color of the TaskGroup node when displayed in the UI
        # ui_fgcolor: Optional[str] = None,
        version: Optional[str] = None,
        enabled: bool = True,
        **kwargs,
    ):
        super().__init__()
        # Cache family tree
        self._family_tree: Optional[Dict[Task, TaskRelatives]] = None
        try:
            self.args: WorkflowArgs = WorkflowArgs(
                name=name,
                tasks=tasks,
                graph=graph,
                on_failure=on_failure,
                run_context=run_context,
                path_context=path_context,
                airflow_context=airflow_context,
                inputs=inputs,
                outputs=outputs,
                pre_checks=pre_checks,
                post_checks=post_checks,
                add_start_task=add_start_task,
                add_end_task=add_end_task,
                use_task_groups=use_task_groups,
                group_id=group_id,
                dag_id=dag_id,
                version=version,
                enabled=enabled,
                **kwargs,
            )
        except Exception as e:
            logger.error(f"Args for {self.name} are not valid")
            raise

    @property
    def tasks(self) -> List[Task]:
        task_list: List[Task] = []
        if self.args and self.args.tasks:
            if isinstance(self.args.tasks, list):
                for task in self.args.tasks:
                    if isinstance(task, list):
                        task_list.extend(task)
                    else:
                        task_list.append(task)
            else:
                logger.warning("Tasks must be a list")
        return task_list

    @tasks.setter
    def tasks(self, tasks: List[Union[Task, List[Task]]]) -> None:
        if self.args is not None and tasks is not None:
            self.args.tasks = tasks

    @property
    def graph(self) -> Optional[Dict[Task, List[Task]]]:
        return self.args.graph if self.args else None

    @graph.setter
    def graph(self, graph: Dict[Task, List[Task]]) -> None:
        if self.args is not None and graph is not None:
            self.args.graph = graph

    @property
    def on_failure(self) -> Literal["stop", "continue"]:
        return self.args.on_failure if self.args else "stop"

    @on_failure.setter
    def on_failure(self, on_failure: Literal["stop", "continue"]) -> None:
        if self.args is not None and on_failure is not None:
            self.args.on_failure = on_failure

    @property
    def run_context(self) -> Optional[RunContext]:
        return self.args.run_context if self.args else None

    @run_context.setter
    def run_context(self, run_context: RunContext) -> None:
        if self.args is not None and run_context is not None:
            self.args.run_context = run_context

    @property
    def path_context(self) -> Optional[PathContext]:
        # Workflow not yet initialized
        if self.args is None:
            return None

        if self.args.path_context is not None:
            # use cached value if available
            return self.args.path_context

        self.args.path_context = build_path_context_from_env()
        return self.args.path_context

    @path_context.setter
    def path_context(self, path_context: PathContext) -> None:
        if self.args is not None and path_context is not None:
            self.args.path_context = path_context

    @property
    def airflow_context(self) -> Optional[AirflowContext]:
        return self.args.airflow_context if self.args else None

    @airflow_context.setter
    def airflow_context(self, airflow_context: AirflowContext) -> None:
        if self.args is not None and airflow_context is not None:
            self.args.airflow_context = airflow_context

    @property
    def inputs(self) -> Optional[List[DataAsset]]:
        return self.args.inputs if self.args else None

    @inputs.setter
    def inputs(self, inputs: List[DataAsset]) -> None:
        if self.args is not None and inputs is not None:
            self.args.inputs = inputs

    @property
    def outputs(self) -> Optional[List[DataAsset]]:
        return self.args.outputs if self.args else None

    @outputs.setter
    def outputs(self, outputs: List[DataAsset]) -> None:
        if self.args is not None and outputs is not None:
            self.args.outputs = outputs

    @property
    def pre_checks(self) -> Optional[List[Check]]:
        return self.args.pre_checks if self.args else None

    @pre_checks.setter
    def pre_checks(self, pre_checks: List[Check]) -> None:
        if self.args is not None and pre_checks is not None:
            self.args.pre_checks = pre_checks

    @property
    def post_checks(self) -> Optional[List[Check]]:
        return self.args.post_checks if self.args else None

    @post_checks.setter
    def post_checks(self, post_checks: List[Check]) -> None:
        if self.args is not None and post_checks is not None:
            self.args.post_checks = post_checks

    @property
    def add_start_task(self) -> bool:
        return self.args.add_start_task if self.args else False

    @add_start_task.setter
    def add_start_task(self, add_start_task: bool) -> None:
        if self.args is not None and add_start_task is not None:
            self.args.add_start_task = add_start_task

    @property
    def add_end_task(self) -> bool:
        return self.args.add_end_task if self.args else False

    @add_end_task.setter
    def add_end_task(self, add_end_task: bool) -> None:
        if self.args is not None and add_end_task is not None:
            self.args.add_end_task = add_end_task

    @property
    def use_task_groups(self) -> bool:
        return self.args.use_task_groups if self.args else False

    @use_task_groups.setter
    def use_task_groups(self, use_task_groups: bool) -> None:
        if self.args is not None and use_task_groups is not None:
            self.args.use_task_groups = use_task_groups

    @property
    def group_id(self) -> str:
        return self.args.group_id if (self.args and self.args.group_id) else self.name

    @group_id.setter
    def group_id(self, group_id: str) -> None:
        if self.args is not None and group_id is not None:
            self.args.group_id = group_id

    @property
    def dag_id(self) -> str:
        return self.args.dag_id if (self.args and self.args.dag_id) else self.name

    @dag_id.setter
    def dag_id(self, dag_id: str) -> None:
        if self.args is not None and dag_id is not None:
            self.args.dag_id = dag_id

    @property
    def description(self) -> Optional[str]:
        return self.args.description if self.args else None

    @description.setter
    def description(self, description: str) -> None:
        if self.args is not None and description is not None:
            self.args.description = description

    @property
    def ui_color(self) -> Optional[str]:
        return self.args.ui_color if self.args else None

    @ui_color.setter
    def ui_color(self, ui_color: str) -> None:
        if self.args is not None and ui_color is not None:
            self.args.ui_color = ui_color

    @property
    def ui_fgcolor(self) -> Optional[str]:
        return self.args.ui_fgcolor if self.args else None

    @ui_fgcolor.setter
    def ui_fgcolor(self, ui_fgcolor: str) -> None:
        if self.args is not None and ui_fgcolor is not None:
            self.args.ui_fgcolor = ui_fgcolor

    ######################################################
    ## Workflow Properties
    ######################################################

    @property
    def airflow_active(self) -> bool:
        return validate_env_vars(self.args.validate_airflow_env)

    @property
    def start_workflow_task_id(self) -> str:
        return f"{self.name}_start"

    @property
    def end_workflow_task_id(self) -> str:
        return f"{self.name}_end"

    @property
    def family_tree(self) -> Optional[Dict[Task, TaskRelatives]]:
        if self.tasks is None:
            return None

        # used cached value if available
        if self._family_tree is not None:
            return self._family_tree

        # build family_tree
        family_tree: Dict[Task, TaskRelatives] = OrderedDict()
        for task in self.tasks:
            family_tree[task] = TaskRelatives()

        if self.graph is None:
            return family_tree

        for task, upstream_list in self.graph.items():
            family_tree[task].add_upstream_list(upstream_list)
            for upstream_task in upstream_list:
                family_tree[upstream_task].add_downstream(task)

        return family_tree

    @property
    def family_tree_dict(self) -> Dict[str, Dict[str, List[str]]]:
        tree: Dict[str, Dict[str, List[str]]] = OrderedDict()
        family_tree = self.family_tree
        if family_tree is None:
            return tree

        for task in family_tree:
            task_tree = {}
            upstream_list = family_tree[task].upstream_list
            downstream_list = family_tree[task].downstream_list
            if upstream_list:
                task_tree["upstream"] = [t.task_id for t in upstream_list]
            if downstream_list:
                task_tree["downstream"] = [t.task_id for t in downstream_list]
            tree[task.task_id] = task_tree
        return tree

    @property
    def roots(self) -> Optional[List[Task]]:
        """
        Roots are tasks that run first. i.e. have nothing upstream of them
        """
        if self.graph is None:
            return self.tasks

        family_tree: Optional[Dict[Task, TaskRelatives]] = self.family_tree
        if family_tree is None:
            return self.tasks

        roots: List[Task] = []
        for task, relatives in family_tree.items():
            if relatives.upstream_list is None:
                roots.append(task)
        return roots

    @property
    def leaves(self) -> Optional[List[Task]]:
        """
        Leaves are tasks that run last. i.e. have nothing downstream of them
        """
        if self.graph is None:
            return self.tasks

        family_tree: Optional[Dict[Task, TaskRelatives]] = self.family_tree
        if family_tree is None:
            return self.tasks

        leaves: List[Task] = []
        for task, relatives in family_tree.items():
            if relatives.downstream_list is None:
                leaves.append(task)
        return leaves

    def __call__(self, *args, **kwargs) -> bool:
        logger.warning("Workflows should not be called directly")
        return True

    ######################################################
    ## Build workflow
    ######################################################

    def build(self) -> bool:
        logger.debug(f"@build not defined for {self.__class__.__name__}")
        return False

    ######################################################
    ## Tasks
    ######################################################

    def add_task(self, task: Task):
        if self.args.tasks is None:
            self.args.tasks = []
        self.args.tasks.append(task)

    ######################################################
    ## Checks
    ######################################################

    def add_pre_check(self, check: Check):
        if self.args.pre_checks is None:
            self.args.pre_checks = []
        self.args.pre_checks.append(check)

    def add_post_check(self, check: Check):
        if self.args.post_checks is None:
            self.args.post_checks = []
        self.args.post_checks.append(check)

    ######################################################
    ## Run workflow
    ######################################################

    def run_task_in_local_env(self, task: Task) -> RunStatus:

        _task_id = task.task_id
        if not task.enabled:
            return RunStatus(name=_task_id, success=True)

        print_info(f"\nRunning task: {_task_id}")
        # Pass down context
        task.run_context = self.run_context
        task.path_context = self.path_context
        task.airflow_context = self.airflow_context
        # Not required for local runs but adding for posterity
        if self.dag_id:
            task.dag_id = self.dag_id
        run_success = task.run_in_local_env()
        if run_success is None:
            logger.warning(
                f"Task must return True/False, returned value: {run_success}"
            )
            run_success = False
        return RunStatus(name=_task_id, success=run_success)

    def run_in_local_env(self, task_id: Optional[str] = None) -> bool:
        """
        Runs a workflow in the local environment where phi wf is called from.

        Returns:
            run_status (bool): True if the run was successful
        """
        from phidata.utils.cli_console import print_run_status_table

        logger.info("--**-- Running Workflow locally")

        if self.tasks is None:
            logger.info("No tasks to run")
            return True

        task_run_status: List[RunStatus] = []
        try:
            if task_id is not None:
                task_id_map = {t.task_id: t for t in self.tasks}
                if task_id in task_id_map:
                    task_run_status.append(
                        self.run_task_in_local_env(task_id_map[task_id])
                    )
                else:
                    logger.error(f"Could not find {task_id} in {task_id_map.keys()}")
            else:
                for task in self.tasks:
                    run_status = self.run_task_in_local_env(task)
                    task_run_status.append(run_status)
                    if not run_status.success and self.args.on_failure == "stop":
                        logger.error(f"Task {task.task_id} failed")
                        break
        except Exception as e:
            logger.error(e)

        print_run_status_table("Task Run Status", task_run_status)
        for run_status in task_run_status:
            if not run_status.success:
                return False
        return True

    def run_task_in_docker_env(
        self,
        task: Task,
        active_container: Any,
        docker_env: Optional[Dict[str, str]] = None,
    ) -> RunStatus:

        _task_id = task.task_id
        if not task.enabled:
            return RunStatus(name=_task_id, success=True)

        print_info(f"\nRunning task: {_task_id}")
        # Pass down context
        task.run_context = self.run_context
        task.path_context = self.path_context
        task.airflow_context = self.airflow_context
        task.dag_id = self.dag_id
        run_success = task.run_in_docker_container(
            active_container=active_container, docker_env=docker_env
        )
        logger.debug(f"Success: {run_success}")
        return RunStatus(name=_task_id, success=run_success)

    def run_in_docker_container(
        self,
        active_container: Any,
        docker_env: Optional[Dict[str, str]] = None,
        task_id: Optional[str] = None,
    ) -> bool:
        """
        Runs a workflow in a docker container.

        Args:
            active_container:
            docker_env:
            task_id:

        Returns:
            run_status (bool): True if the run was successful

        Notes:
            * This function runs in the local environment where phi wf is called from.
            But executes `airflow` commands in the docker container to run the workflow
            * For the airflow tasks to be available, they need to be added to the workflow DAG
            using add_airflow_tasks_to_dag()
        """
        from phidata.utils.cli_console import print_run_status_table

        logger.info("--**-- Running Workflow in docker container")

        if self.tasks is None:
            logger.info("No tasks to run")
            return True

        task_run_status: List[RunStatus] = []
        try:
            if task_id is not None:
                task_id_map = {t.task_id: t for t in self.tasks}
                if task_id in task_id_map:
                    task_run_status.append(
                        self.run_task_in_docker_env(
                            task=task_id_map[task_id],
                            active_container=active_container,
                            docker_env=docker_env,
                        )
                    )
                else:
                    logger.error(f"Could not find {task_id} in {task_id_map.keys()}")
            else:
                for task in self.tasks:
                    run_status = self.run_task_in_docker_env(
                        task=task,
                        active_container=active_container,
                        docker_env=docker_env,
                    )
                    task_run_status.append(run_status)
                    if not run_status.success and self.args.on_failure == "stop":
                        raise TaskFailure(f"Task {task.task_id} failed")
        except TaskFailure as tf:
            logger.error(tf)

        print_run_status_table("Task Run Status", task_run_status)
        for run_status in task_run_status:
            if not run_status.success:
                return False
        return True

    def run_in_k8s_container(
        self,
        pod: Any,
        k8s_api_client: Any,
        container_name: Optional[str] = None,
        k8s_env: Optional[Dict[str, str]] = None,
    ) -> bool:
        logger.info("--**-- Running Workflow in k8s container")

        if self.tasks is None:
            logger.info("No tasks to run")
            return True

        task_run_status: List[RunStatus] = []
        for idx, task in enumerate(self.tasks, start=1):
            task_name = task.name or "{}__{}".format(task.__class__.__name__, idx)
            print_subheading(f"\nRunning {task_name}")
            # Pass down context
            task.run_context = self.run_context
            task.path_context = self.path_context
            task.airflow_context = self.airflow_context
            task.dag_id = self.dag_id
            run_success = task.run_in_k8s_container(
                pod=pod,
                k8s_api_client=k8s_api_client,
                container_name=container_name,
                k8s_env=k8s_env,
            )
            task_run_status.append(RunStatus(name=task_name, success=run_success))

        print_subheading("\nTask run status:")
        print_info(
            "\n".join(
                [
                    "{}: {}".format(task.name, "Success" if task.success else "Fail")
                    for task in task_run_status
                ]
            )
        )
        print_info("")
        for run_status in task_run_status:
            if not run_status.success:
                return False
        return True

    ######################################################
    ## New Run functions
    ######################################################

    def run_local(
        self,
        env: Optional[Dict[str, str]] = None,
    ) -> bool:
        """
        Runs the workflow in the local environment.

        Args:
            env:

        Returns:
            run_status (bool): True if the run was successful
        """
        logger.debug(f"@run_local not defined for {self.name}")
        return False

    def run_docker(
        self,
        env: Optional[Dict[str, str]] = None,
        context: Optional[DockerContext] = None,
    ) -> bool:
        """
        Runs the workflow in a docker container.

        Args:
            env:
            context:

        Returns:
            run_status (bool): True if the run was successful
        """
        logger.debug(f"@run_docker not defined for {self.name}")
        return False

    def run_k8s(
        self,
        env: Optional[Dict[str, str]] = None,
        context: Optional[K8sContext] = None,
    ) -> bool:
        """
        Runs the workflow in a k8s pod.

        Args:
            env:
            context:

        Returns:
            run_status (bool): True if the run was successful
        """
        logger.debug(f"@run_k8s not defined for {self.name}")
        return False

    ######################################################
    ## Airflow functions
    ######################################################

    # def get_task_group(self, dag: Optional[Any]) -> Optional[Any]:
    #
    #     from airflow.utils.task_group import TaskGroup
    #
    #     # check group_id
    #     group_id = self.group_id
    #     if group_id is None:
    #         logger.error("Workflow group_id unavailable")
    #         return False
    #
    #     # build a dict of not_null_args which are passed to the TaskGroup()
    #     not_null_args: Dict[str, Any] = {}
    #     if self.description is not None:
    #         not_null_args["tooltip"] = self.description
    #     if self.ui_color is not None:
    #         not_null_args["ui_color"] = self.ui_color
    #     if self.ui_fgcolor is not None:
    #         not_null_args["ui_fgcolor"] = self.ui_fgcolor
    #
    #     return TaskGroup(
    #         group_id=group_id, dag=dag, add_suffix_on_collision=True, **not_null_args
    #     )

    def get_start_workflow_task(self, dag: Optional[Any]) -> Any:

        from phidata.airflow.operators.empty import EmptyOperator

        return EmptyOperator(task_id=self.start_workflow_task_id, dag=dag)

    def get_end_workflow_task(self, dag: Optional[Any]) -> Any:

        from phidata.airflow.operators.empty import EmptyOperator

        return EmptyOperator(task_id=self.end_workflow_task_id, dag=dag)

    def add_airflow_tasks_to_dag(self, dag: Any) -> bool:
        """
        Add tasks to the airflow DAG.

        Args:
            dag:

        Returns:

        Notes:
            * This function is called as part of the create_airflow_dag() function
        """
        # Skips DAG creation on local machines
        if not self.airflow_active:
            return False

        from airflow import DAG
        from airflow.models import BaseOperator

        # from airflow.utils.task_group import TaskGroup
        from phidata.airflow.operators.empty import EmptyOperator

        # Cast DAG to correct type
        dag = cast(DAG, dag)

        # Check tasks
        if self.tasks is None:
            logger.error(f"Workflow {self.name} has no tasks")
            return False

        # If self.use_task_groups = True, then the task_group variable is
        # updated using the self.get_task_group() function
        # This way, when we provide the task_group parameter to each task,
        #   the task gets automatically added to the task_group()
        #   if the task_group variable is not None
        # task_group: Optional[TaskGroup] = None
        # if self.use_task_groups:
        #     task_group = self.get_task_group(dag=dag)
        #     if task_group is None or not isinstance(task_group, TaskGroup):
        #         logger.error(f"Could not create task_group: {self.group_id}")
        #         return False

        # Build airflow tasks for this workflow
        airflow_tasks: Dict[str, BaseOperator] = OrderedDict()
        for task in self.tasks:
            task_id = task.task_id
            # Pass down context
            task.run_context = self.run_context
            task.path_context = self.path_context
            task.airflow_context = self.airflow_context

            airflow_task_operator = task.add_airflow_task(dag=dag)
            if airflow_task_operator is None or not isinstance(
                airflow_task_operator, BaseOperator
            ):
                logger.warning(f"Task: {task_id} could not be created")
                continue

            airflow_tasks[task_id] = airflow_task_operator

        # Set dependencies
        if self.graph is not None:
            # logger.info("Updating dependencies")
            for downstream_task, upstream_task_list in self.graph.items():
                # logger.info(f"Task: {downstream_task.task_id}")
                downstream_airflow_task = airflow_tasks[downstream_task.task_id]
                upstream_airflow_task_list = [
                    airflow_tasks[task.task_id] for task in upstream_task_list
                ]
                downstream_airflow_task.set_upstream(upstream_airflow_task_list)
                # logger.info(
                #     f"Upstream tasks: {downstream_airflow_task.upstream_task_ids}"
                # )

        # Add workflow_start task if self.add_start_task
        if self.add_start_task:
            workflow_start: EmptyOperator = self.get_start_workflow_task(dag=dag)
            roots = self.roots
            # Add workflow_end to be upstream of all root tasks
            if roots is not None:
                for root_task in roots:
                    root_task_airflow_op = airflow_tasks[root_task.task_id]
                    # Add root task to be downstream of workflow_start
                    # logger.info(
                    #     f"Adding {root_task.task_id} downstream to {workflow_start.task_id}"
                    # )
                    workflow_start.set_downstream(root_task_airflow_op)

        # Add workflow_end task if self.add_end_task
        if self.add_end_task:
            workflow_end: EmptyOperator = self.get_end_workflow_task(dag=dag)
            leaves = self.leaves
            # Add workflow_end to be downstream of all leaf tasks
            if leaves is not None:
                for leaf_task in leaves:
                    leaf_task_airflow_op = airflow_tasks[leaf_task.task_id]
                    # Add workflow_end to be downstream of leaf task
                    # logger.info(
                    #     f"Adding {workflow_end.task_id} downstream to {leaf_task_airflow_op.task_id}"
                    # )
                    leaf_task_airflow_op.set_downstream(workflow_end)

        return True

    def create_airflow_dag(
        self,
        # The id of the DAG; must consist exclusively of
        # alphanumeric characters, dashes, dots and underscores (all ASCII)
        dag_id: Optional[str] = None,
        owner: Optional[str] = "airflow",
        depends_on_past: Optional[bool] = False,
        # The description for the DAG to e.g. be shown on the webserver
        description: Optional[str] = None,
        # Defines how often that DAG runs, this
        #  timedelta object gets added to your latest task instance's
        #  execution_date to figure out the next schedule
        # Default: daily
        schedule_interval: Optional[Union[str, timedelta]] = "@daily",
        # The timestamp from which the scheduler will
        #  attempt to backfill
        start_date: Optional[datetime] = None,
        # if start_date is not provided, use start_days_ago
        start_days_ago: int = 1,
        # A date beyond which your DAG won't run, leave to None
        # for open-ended scheduling
        end_date: Optional[datetime] = None,
        # This list of folders (non-relative)
        #   defines where jinja will look for your templates. Order matters.
        #   Note that jinja/airflow includes the path of your DAG file by default
        template_searchpath: Optional[Union[str, Iterable[str]]] = None,
        # a dictionary of macros that will be exposed
        #  in your jinja templates. For example, passing ``dict(foo='bar')``
        #  to this argument allows you to ``{{ foo }}`` in all jinja
        #  templates related to this DAG. Note that you can pass any
        #  type of object here.
        user_defined_macros: Optional[Dict] = None,
        # a dictionary of filters that will be exposed
        #  in your jinja templates. For example, passing
        #  ``dict(hello=lambda name: 'Hello %s' % name)`` to this argument allows
        #  you to ``{{ 'world' | hello }}`` in all jinja templates related to
        #  this DAG.
        user_defined_filters: Optional[Dict] = None,
        # A dictionary of default parameters to be used
        #  as constructor keyword parameters when initialising operators.
        #  Note that operators have the same hook, and precede those defined
        #  here, meaning that if your dict contains `'depends_on_past': True`
        #  here and `'depends_on_past': False` in the operator's call
        #  `default_args`, the actual value will be `False`.
        default_args: Optional[Dict] = None,
        # the number of task instances allowed to run concurrently
        max_active_tasks: int = 32,
        # maximum number of active DAG runs, beyond this
        #  number of DAG runs in a running state, the scheduler won't create
        #  new active DAG runs
        max_active_runs: int = 8,
        # specify how long a DagRun should be up before
        # timing out / failing, so that new DagRuns can be created. The timeout
        # is only enforced for scheduled DagRuns.
        dagrun_timeout: Optional[timedelta] = None,
        # Specify DAG default view (grid, graph, duration,
        # gantt, landing_times), default grid
        # default_view: Optional[str] = None,
        # Specify DAG orientation in graph view (LR, TB, RL, BT), default LR
        # orientation: Optional[str] = None,
        # Perform scheduler catchup (or only run latest)? Defaults to True
        catchup: bool = True,
        doc_md: Optional[str] = None,
        # a dictionary of DAG level parameters that are made
        # accessible in templates, namespaced under `params`. These
        # params can be overridden at the task level.
        params: Optional[Dict] = None,
        # Specify optional DAG-level actions, e.g.,
        # "{'role1': {'can_read'}, 'role2': {'can_read', 'can_edit', 'can_delete'}}"
        access_control: Optional[Dict] = None,
        # Specifies if the dag is paused when created for the first time.
        # If the dag exists already, this flag will be ignored. If this optional parameter
        # is not specified, the global config setting will be used.
        is_paused_upon_creation: Optional[bool] = None,
        # additional configuration options to be passed to Jinja
        # ``Environment`` for template rendering
        # **Example**: to avoid Jinja from removing a trailing newline from template strings ::
        #
        #     DAG(dag_id='my-dag',
        #         jinja_environment_kwargs={
        #             'keep_trailing_newline': True,
        #             # some other jinja2 Environment options here
        #         }
        #     )
        #
        # **See**: `Jinja Environment documentation
        # <https://jinja.palletsprojects.com/en/2.11.x/api/#jinja2.Environment>`_
        jinja_environment_kwargs: Optional[Dict] = None,
        # If True, uses a Jinja ``NativeEnvironment``
        # to render templates as native Python types. If False, a Jinja
        # ``Environment`` is used to render templates as string values.
        render_template_as_native_obj: bool = False,
        tags: Optional[List[str]] = None,
    ) -> Any:
        """
        Create Airflow DAG for Workflow.
        """

        # Skips DAG creation on local machines
        if not self.airflow_active:
            return None

        from airflow import DAG

        operator_default_args = default_args
        if operator_default_args is None:
            operator_default_args = {}
        operator_default_args.update(
            {
                "owner": owner,
                "depends_on_past": depends_on_past,
            }
        )
        if start_date is None:
            from airflow.utils.dates import days_ago

            start_date = days_ago(start_days_ago)

        # dag_id order of preference:
        #   - use dag_id provided to this function
        #   - use the dag_id param of the workflow
        #   - use the name param of the workflow
        dag_id = dag_id or self.dag_id
        if dag_id is None:
            logger.error("Workflow dag_id unavailable")
            return None

        dag = DAG(
            dag_id=dag_id,
            description=description,
            schedule_interval=schedule_interval,
            start_date=start_date,
            end_date=end_date,
            template_searchpath=template_searchpath,
            user_defined_macros=user_defined_macros,
            user_defined_filters=user_defined_filters,
            default_args=operator_default_args,
            max_active_tasks=max_active_tasks,
            max_active_runs=max_active_runs,
            dagrun_timeout=dagrun_timeout,
            catchup=catchup,
            doc_md=doc_md,
            params=params,
            access_control=access_control,
            is_paused_upon_creation=is_paused_upon_creation,
            jinja_environment_kwargs=jinja_environment_kwargs,
            render_template_as_native_obj=render_template_as_native_obj,
            tags=tags,
        )
        add_task_success = self.add_airflow_tasks_to_dag(dag=dag)
        if not add_task_success:
            logger.error(f"Tasks for workflow {self.name} could not be added")
        logger.debug(f"Airflow dag {dag_id} created")
        return dag

    def task(
        self,
        func: Optional[PythonTaskType] = None,
        *,
        name: Optional[str] = None,
        task_id: Optional[str] = None,
        version: Optional[str] = None,
        enabled: bool = True,
        args: Optional[Tuple[Any, ...]] = None,
        **kwargs,
    ) -> Callable[..., PythonTask]:
        """
        Adds a PythonTask to this workflow

        Usage:

        """
        from functools import partial

        def decorator(f: PythonTaskType) -> PythonTask:
            # logger.info(f"workflow.task func: {func}")
            # logger.info(f"workflow.task f: {f}")
            # logger.info(f"workflow.task args: {args}")
            # logger.info(f"workflow.task kwargs: {kwargs}")
            task_name = name or f.__name__
            try:
                # create a partial function that can be called by the PythonTask
                entrypoint = (
                    partial(f, *args, **kwargs)
                    if args is not None
                    else partial(f, **kwargs)
                )
                python_task = PythonTask(
                    name=task_name,
                    entrypoint=entrypoint,
                    task_id=task_id,
                    dag_id=self.dag_id,
                    entrypoint_args=args,
                    entrypoint_kwargs=kwargs,
                    version=version,
                    enabled=enabled,
                )
                self.add_task(task=python_task)
                return python_task
            except Exception as e:
                logger.error(f"Could not create task: {task_name}")
                logger.error(e)
                raise

        return decorator



from typing import Optional, Any


class InfraApiClient:
    def __init__(self) -> None:
        self._api_client: Optional[Any] = None

    def is_initialized(self) -> bool:
        """
        Returns True if the ApiClient subclass is initialized properly.
        Needs to by implemented by each subclass
        """
        return False



from phidata.infra.k8s.create.apiextensions_k8s_io.v1.custom_resource_definition import (
    CreateCustomResourceDefinition,
    CustomResourceDefinitionNames,
    CustomResourceDefinitionVersion,
    V1JSONSchemaProps,
)

######################################################
## Traefik CRDs
######################################################
traefik_name = "traefik"
ingressroute_crd = CreateCustomResourceDefinition(
    crd_name="ingressroutes.traefik.containo.us",
    app_name=traefik_name,
    group="traefik.containo.us",
    names=CustomResourceDefinitionNames(
        kind="IngressRoute",
        list_kind="IngressRouteList",
        plural="ingressroutes",
        singular="ingressroute",
    ),
    annotations={
        "controller-gen.kubebuilder.io/version": "v0.6.2",
    },
    versions=[
        CustomResourceDefinitionVersion(
            name="v1alpha1",
            served=True,
            storage=True,
            open_apiv3_schema=V1JSONSchemaProps(
                description="IngressRoute is an Ingress CRD specification.",
                type="object",
                required=["metadata", "spec"],
                properties={
                    "apiVersion": V1JSONSchemaProps(
                        description="APIVersion defines the versioned schema of this representation of an object. Servers should convert recognized schemas to the latest internal value, and may reject unrecognized values. More info: https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#resources",
                        type="string",
                    ),
                    "kind": V1JSONSchemaProps(
                        description="Kind is a string value representing the REST resource this object represents. Servers may infer this from the endpoint the client submits requests to. Cannot be updated. In CamelCase. More info: https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#types-kinds",
                        type="string",
                    ),
                    "metadata": V1JSONSchemaProps(type="object"),
                    "spec": V1JSONSchemaProps(
                        description="IngressRouteSpec is a specification for a IngressRouteSpec resource.",
                        type="object",
                        required=["routes"],
                        properties={
                            "entryPoints": V1JSONSchemaProps(
                                type="array",
                                items={
                                    "type": "string",
                                },
                            ),
                            "routes": V1JSONSchemaProps(
                                type="array",
                                items={
                                    "description": "Route contains the set of routes.",
                                    "type": "object",
                                    "required": ["kind", "match"],
                                    "properties": {
                                        "kind": V1JSONSchemaProps(
                                            type="string", enum=["Rule"]
                                        ),
                                        "match": V1JSONSchemaProps(
                                            type="string",
                                        ),
                                        "middlewares": V1JSONSchemaProps(
                                            type="array",
                                            items={
                                                "description": "Route contains the set of routes.",
                                                "type": "object",
                                                "required": ["name"],
                                                "properties": {
                                                    "name": V1JSONSchemaProps(
                                                        type="string",
                                                    ),
                                                    "namespace": V1JSONSchemaProps(
                                                        type="string",
                                                    ),
                                                },
                                            },
                                        ),
                                        "priority": V1JSONSchemaProps(
                                            type="integer",
                                        ),
                                        "services": V1JSONSchemaProps(
                                            type="array",
                                            items={
                                                "description": "Service defines an upstream to proxy traffic.",
                                                "type": "object",
                                                "required": ["name"],
                                                "properties": {
                                                    "kind": V1JSONSchemaProps(
                                                        type="string",
                                                        enum=[
                                                            "Service",
                                                            "TraefikService",
                                                        ],
                                                    ),
                                                    "name": V1JSONSchemaProps(
                                                        description="Name is a reference to a Kubernetes Service object (for a load-balancer of servers), or to a TraefikServic                               object (service load-balancer, mirroring, etc). The differentiation between the two is specified in the Kind field.",
                                                        type="string",
                                                    ),
                                                    "passHostHeader": V1JSONSchemaProps(
                                                        type="boolean",
                                                    ),
                                                    "port": V1JSONSchemaProps(
                                                        any_of=[
                                                            V1JSONSchemaProps(
                                                                type="integer",
                                                            ),
                                                            V1JSONSchemaProps(
                                                                type="string",
                                                            ),
                                                        ],
                                                        x_kubernetes_int_or_string=True,
                                                    ),
                                                    "responseForwarding": V1JSONSchemaProps(
                                                        description="ResponseForwarding holds configuration for the forward of the response.",
                                                        type="object",
                                                        properties={
                                                            "flushInterval": V1JSONSchemaProps(
                                                                type="string",
                                                            )
                                                        },
                                                    ),
                                                    "scheme": V1JSONSchemaProps(
                                                        type="string",
                                                    ),
                                                    "serversTransport": V1JSONSchemaProps(
                                                        type="string",
                                                    ),
                                                    "sticky": V1JSONSchemaProps(
                                                        description="Sticky holds the sticky configuration.",
                                                        type="object",
                                                        properties={
                                                            "cookie": V1JSONSchemaProps(
                                                                description="Cookie holds the sticky configuration based on cookie",
                                                                type="object",
                                                                properties={
                                                                    "httpOnly": V1JSONSchemaProps(
                                                                        type="boolean",
                                                                    ),
                                                                    "name": V1JSONSchemaProps(
                                                                        type="string",
                                                                    ),
                                                                    "sameSite": V1JSONSchemaProps(
                                                                        type="string",
                                                                    ),
                                                                    "secure": V1JSONSchemaProps(
                                                                        type="boolean",
                                                                    ),
                                                                },
                                                            )
                                                        },
                                                    ),
                                                    "strategy": V1JSONSchemaProps(
                                                        type="string",
                                                    ),
                                                    "weight": V1JSONSchemaProps(
                                                        description="Weight should only be specified when Name references a TraefikService object (and to be precise, one that embeds a Weighted Round Robin).",
                                                        type="integer",
                                                    ),
                                                },
                                            },
                                        ),
                                    },
                                },
                            ),
                            "tls": V1JSONSchemaProps(
                                description="TLS contains the TLS certificates configuration of the routes. To enable Let's Encrypt, use an empty TLS struct, e.g. in YAML: \n \t tls: {} # inline format \n \t tls: \t   secretName: # block format",
                                type="object",
                                properties={
                                    "certResolver": V1JSONSchemaProps(
                                        type="string",
                                    ),
                                    "domains": V1JSONSchemaProps(
                                        type="array",
                                        items={
                                            "description": "Domain holds a domain name with SANs.",
                                            "type": "object",
                                            "properties": {
                                                "main": V1JSONSchemaProps(
                                                    type="string",
                                                ),
                                                "sans": V1JSONSchemaProps(
                                                    type="array",
                                                    items={
                                                        "type": "string",
                                                    },
                                                ),
                                            },
                                        },
                                    ),
                                    "options": V1JSONSchemaProps(
                                        description="Options is a reference to a TLSOption, that specifies the parameters of the TLS connection.",
                                        type="object",
                                        required=["name"],
                                        properties={
                                            "name": V1JSONSchemaProps(
                                                type="string",
                                            ),
                                            "namespace": V1JSONSchemaProps(
                                                type="string",
                                            ),
                                        },
                                    ),
                                    "secretName": V1JSONSchemaProps(
                                        description="SecretName is the name of the referenced Kubernetes Secret to specify the certificate details.",
                                        type="string",
                                    ),
                                    "store": V1JSONSchemaProps(
                                        description="Store is a reference to a TLSStore, that specifies the parameters of the TLS store.",
                                        type="object",
                                        required=["name"],
                                        properties={
                                            "name": V1JSONSchemaProps(
                                                type="string",
                                            ),
                                            "namespace": V1JSONSchemaProps(
                                                type="string",
                                            ),
                                        },
                                    ),
                                },
                            ),
                        },
                    ),
                },
            ),
        )
    ],
)

ingressroutetcp_crd = CreateCustomResourceDefinition(
    crd_name="ingressroutetcps.traefik.containo.us",
    app_name=traefik_name,
    group="traefik.containo.us",
    names=CustomResourceDefinitionNames(
        kind="IngressRouteTCP",
        list_kind="IngressRouteTCPList",
        plural="ingressroutetcps",
        singular="ingressroutetcp",
    ),
    annotations={
        "controller-gen.kubebuilder.io/version": "v0.6.2",
    },
    versions=[
        CustomResourceDefinitionVersion(
            name="v1alpha1",
            open_apiv3_schema=V1JSONSchemaProps(
                description="IngressRouteTCP is an Ingress CRD specification.",
                type="object",
                required=["metadata", "spec"],
                properties={
                    "apiVersion": V1JSONSchemaProps(
                        description="APIVersion defines the versioned schema of this representation of an object. Servers should convert recognized schemas to the latest internal value, and may reject unrecognized values. More info: https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#resources",
                        type="string",
                    ),
                    "kind": V1JSONSchemaProps(
                        description="Kind is a string value representing the REST resource this object represents. Servers may infer this from the endpoint the client submits requests to. Cannot be updated. In CamelCase. More info: https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#types-kinds",
                        type="string",
                    ),
                    "metadata": V1JSONSchemaProps(type="object"),
                    "spec": V1JSONSchemaProps(
                        description="IngressRouteTCPSpec is a specification for a IngressRouteTCPSpec resource.",
                        type="object",
                        required=["routes"],
                        properties={
                            "entryPoints": V1JSONSchemaProps(
                                type="array",
                                items={
                                    "type": "string",
                                },
                            ),
                            "routes": V1JSONSchemaProps(
                                type="array",
                                items={
                                    "description": "RouteTCP contains the set of routes.",
                                    "type": "object",
                                    "required": ["match"],
                                    "properties": {
                                        "match": V1JSONSchemaProps(
                                            type="string",
                                        ),
                                        "middlewares": V1JSONSchemaProps(
                                            description="Middlewares contains references to MiddlewareTCP resources.",
                                            type="array",
                                            items={
                                                "description": "ObjectReference is a generic reference to a Traefik resource.",
                                                "type": "object",
                                                "required": ["name"],
                                                "properties": {
                                                    "name": V1JSONSchemaProps(
                                                        type="string",
                                                    ),
                                                    "namespace": V1JSONSchemaProps(
                                                        type="string",
                                                    ),
                                                },
                                            },
                                        ),
                                        "services": V1JSONSchemaProps(
                                            type="array",
                                            items={
                                                "description": "ServiceTCP defines an upstream to proxy traffic.",
                                                "type": "object",
                                                "required": ["name", "port"],
                                                "properties": {
                                                    "name": V1JSONSchemaProps(
                                                        type="string",
                                                    ),
                                                    "namespace": V1JSONSchemaProps(
                                                        type="string",
                                                    ),
                                                    "port": V1JSONSchemaProps(
                                                        any_of=[
                                                            V1JSONSchemaProps(
                                                                type="integer",
                                                            ),
                                                            V1JSONSchemaProps(
                                                                type="string",
                                                            ),
                                                        ],
                                                        x_kubernetes_int_or_string=True,
                                                    ),
                                                    "proxyProtocol": V1JSONSchemaProps(
                                                        description="ProxyProtocol holds the ProxyProtocol configuration.",
                                                        type="object",
                                                        properties={
                                                            "version": V1JSONSchemaProps(
                                                                type="integer",
                                                            )
                                                        },
                                                    ),
                                                    "terminationDelay": V1JSONSchemaProps(
                                                        type="integer",
                                                    ),
                                                    "weight": V1JSONSchemaProps(
                                                        type="integer",
                                                    ),
                                                },
                                            },
                                        ),
                                    },
                                },
                            ),
                            "tls": V1JSONSchemaProps(
                                description="TLSTCP contains the TLS certificates configuration of the routes. To enable Let's Encrypt, use an empty TLS struct, e.g. in YAML: \n \t tls: {} # inline format \n \t tls: \t   secretName: # block format",
                                type="object",
                                properties={
                                    "certResolver": V1JSONSchemaProps(
                                        type="string",
                                    ),
                                    "domains": V1JSONSchemaProps(
                                        type="array",
                                        items={
                                            "description": "Domain holds a domain name with SANs.",
                                            "type": "object",
                                            "properties": {
                                                "main": V1JSONSchemaProps(
                                                    type="string",
                                                ),
                                                "sans": V1JSONSchemaProps(
                                                    type="array",
                                                    items={
                                                        "type": "string",
                                                    },
                                                ),
                                            },
                                        },
                                    ),
                                    "options": V1JSONSchemaProps(
                                        description="Options is a reference to a TLSOption, that specifies the parameters of the TLS connection.",
                                        type="object",
                                        required=["name"],
                                        properties={
                                            "name": V1JSONSchemaProps(
                                                type="string",
                                            ),
                                            "namespace": V1JSONSchemaProps(
                                                type="string",
                                            ),
                                        },
                                    ),
                                    "passthrough": V1JSONSchemaProps(
                                        type="boolean",
                                    ),
                                    "secretName": V1JSONSchemaProps(
                                        description="SecretName is the name of the referenced Kubernetes Secret to specify the certificate details.",
                                        type="string",
                                    ),
                                    "store": V1JSONSchemaProps(
                                        description="Store is a reference to a TLSStore, that specifies the parameters of the TLS store.",
                                        type="object",
                                        required=["name"],
                                        properties={
                                            "name": V1JSONSchemaProps(
                                                type="string",
                                            ),
                                            "namespace": V1JSONSchemaProps(
                                                type="string",
                                            ),
                                        },
                                    ),
                                },
                            ),
                        },
                    ),
                },
            ),
        )
    ],
)

ingressrouteudp_crd = CreateCustomResourceDefinition(
    crd_name="ingressrouteudps.traefik.containo.us",
    app_name=traefik_name,
    group="traefik.containo.us",
    names=CustomResourceDefinitionNames(
        kind="IngressRouteUDP",
        list_kind="IngressRouteUDPList",
        plural="ingressrouteudps",
        singular="ingressrouteudp",
    ),
    annotations={
        "controller-gen.kubebuilder.io/version": "v0.6.2",
    },
    versions=[
        CustomResourceDefinitionVersion(
            name="v1alpha1",
            open_apiv3_schema=V1JSONSchemaProps(
                description="IngressRouteUDP is an Ingress CRD specification.",
                type="object",
                required=["metadata", "spec"],
                properties={
                    "apiVersion": V1JSONSchemaProps(
                        description="APIVersion defines the versioned schema of this representation of an object. Servers should convert recognized schemas to the latest internal value, and may reject unrecognized values. More info: https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#resources",
                        type="string",
                    ),
                    "kind": V1JSONSchemaProps(
                        description="Kind is a string value representing the REST resource this object represents. Servers may infer this from the endpoint the client submits requests to. Cannot be updated. In CamelCase. More info: https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#types-kinds",
                        type="string",
                    ),
                    "metadata": V1JSONSchemaProps(type="object"),
                    "spec": V1JSONSchemaProps(
                        description="IngressRouteUDPSpec is a specification for a IngressRouteUDPSpec resource.",
                        type="object",
                        required=["routes"],
                        properties={
                            "entryPoints": V1JSONSchemaProps(
                                type="array",
                                items={
                                    "type": "string",
                                },
                            ),
                            "routes": V1JSONSchemaProps(
                                type="array",
                                items={
                                    "description": "RouteUDP contains the set of routes.",
                                    "type": "object",
                                    "properties": {
                                        "services": V1JSONSchemaProps(
                                            type="array",
                                            items={
                                                "description": "ServiceUDP defines an upstream to proxy traffic.",
                                                "type": "object",
                                                "required": ["name", "port"],
                                                "properties": {
                                                    "name": V1JSONSchemaProps(
                                                        type="string",
                                                    ),
                                                    "namespace": V1JSONSchemaProps(
                                                        type="string",
                                                    ),
                                                    "port": V1JSONSchemaProps(
                                                        any_of=[
                                                            V1JSONSchemaProps(
                                                                type="integer",
                                                            ),
                                                            V1JSONSchemaProps(
                                                                type="string",
                                                            ),
                                                        ],
                                                        x_kubernetes_int_or_string=True,
                                                    ),
                                                    "weight": V1JSONSchemaProps(
                                                        type="integer",
                                                    ),
                                                },
                                            },
                                        ),
                                    },
                                },
                            ),
                        },
                    ),
                },
            ),
        )
    ],
)

middleware_crd = CreateCustomResourceDefinition(
    crd_name="middlewares.traefik.containo.us",
    app_name=traefik_name,
    group="traefik.containo.us",
    names=CustomResourceDefinitionNames(
        kind="Middleware",
        list_kind="MiddlewareList",
        plural="middlewares",
        singular="middleware",
    ),
    annotations={
        "controller-gen.kubebuilder.io/version": "v0.6.2",
    },
    versions=[
        CustomResourceDefinitionVersion(
            name="v1alpha1",
            open_apiv3_schema=V1JSONSchemaProps(
                description="Middleware is a specification for a Middleware resource.",
                type="object",
                required=["metadata", "spec"],
                properties={
                    "apiVersion": V1JSONSchemaProps(
                        description="APIVersion defines the versioned schema of this representation of an object. Servers should convert recognized schemas to the latest internal value, and may reject unrecognized values. More info: https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#resources",
                        type="string",
                    ),
                    "kind": V1JSONSchemaProps(
                        description="Kind is a string value representing the REST resource this object represents. Servers may infer this from the endpoint the client submits requests to. Cannot be updated. In CamelCase. More info: https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#types-kinds",
                        type="string",
                    ),
                    "metadata": V1JSONSchemaProps(type="object"),
                    "spec": V1JSONSchemaProps(
                        description="MiddlewareSpec holds the Middleware configuration.",
                        type="object",
                        properties={
                            "addPrefix": V1JSONSchemaProps(
                                description="AddPrefix holds the AddPrefix configuration.",
                                type="object",
                                properties={
                                    "prefix": V1JSONSchemaProps(
                                        type="string",
                                    ),
                                },
                            ),
                            "basicAuth": V1JSONSchemaProps(
                                description="BasicAuth holds the HTTP basic authentication configuration.",
                                type="object",
                                properties={
                                    "headerField": V1JSONSchemaProps(
                                        type="string",
                                    ),
                                    "realm": V1JSONSchemaProps(
                                        type="string",
                                    ),
                                    "removeHeader": V1JSONSchemaProps(
                                        type="boolean",
                                    ),
                                    "secret": V1JSONSchemaProps(
                                        type="string",
                                    ),
                                },
                            ),
                            "buffering": V1JSONSchemaProps(
                                description="Buffering holds the request/response buffering configuration.",
                                type="object",
                                properties={
                                    "maxRequestBodyBytes": V1JSONSchemaProps(
                                        format="int64",
                                        type="integer",
                                    ),
                                    "maxResponseBodyBytes": V1JSONSchemaProps(
                                        format="int64",
                                        type="integer",
                                    ),
                                    "memRequestBodyBytes": V1JSONSchemaProps(
                                        format="int64",
                                        type="integer",
                                    ),
                                    "memResponseBodyBytes": V1JSONSchemaProps(
                                        format="int64",
                                        type="integer",
                                    ),
                                    "retryExpression": V1JSONSchemaProps(
                                        type="string",
                                    ),
                                },
                            ),
                            "chain": V1JSONSchemaProps(
                                description="Chain holds a chain of middlewares.",
                                type="object",
                                properties={
                                    "middlewares": V1JSONSchemaProps(
                                        type="array",
                                        items={
                                            "description": "MiddlewareRef is a ref to the Middleware resources.",
                                            "type": "object",
                                            "required": ["name"],
                                            "properties": {
                                                "name": V1JSONSchemaProps(
                                                    type="string",
                                                ),
                                                "namespace": V1JSONSchemaProps(
                                                    type="string",
                                                ),
                                            },
                                        },
                                    ),
                                },
                            ),
                            "circuitBreaker": V1JSONSchemaProps(
                                description="CircuitBreaker holds the circuit breaker configuration.",
                                type="object",
                                properties={
                                    "expression": V1JSONSchemaProps(
                                        type="string",
                                    ),
                                },
                            ),
                            "compress": V1JSONSchemaProps(
                                description="Compress holds the compress configuration.",
                                type="object",
                                properties={
                                    "excludedContentTypes": V1JSONSchemaProps(
                                        type="array",
                                        items={
                                            "type": "string",
                                        },
                                    ),
                                    "minResponseBodyBytes": V1JSONSchemaProps(
                                        type="integer",
                                    ),
                                },
                            ),
                            "contentType": V1JSONSchemaProps(
                                description="ContentType middleware - or rather its unique `autoDetect` option - specifies whether to let the `Content-Type` header, if it has not been set by the backend, be automatically set to a value derived from the contents of the response. As a proxy, the default behavior should be to leave the header alone, regardless of what the backend did with it. However, the historic default was to always auto-detect and set the header if it was nil, and it is going to be kept that way in order to support users currently relying on it. This middleware exists to enable the correct behavior until at least the default one can be changed in a future version.",
                                type="object",
                                properties={
                                    "autoDetect": V1JSONSchemaProps(
                                        type="boolean",
                                    ),
                                },
                            ),
                            "digestAuth": V1JSONSchemaProps(
                                description="DigestAuth holds the Digest HTTP authentication configuration.",
                                type="object",
                                properties={
                                    "headerField": V1JSONSchemaProps(
                                        type="string",
                                    ),
                                    "realm": V1JSONSchemaProps(
                                        type="string",
                                    ),
                                    "removeHeader": V1JSONSchemaProps(
                                        type="boolean",
                                    ),
                                    "secret": V1JSONSchemaProps(
                                        type="string",
                                    ),
                                },
                            ),
                            "errors": V1JSONSchemaProps(
                                description="ErrorPage holds the custom error page configuration.",
                                type="object",
                                properties={
                                    "query": V1JSONSchemaProps(
                                        type="string",
                                    ),
                                    "service": V1JSONSchemaProps(
                                        description="Service defines an upstream to proxy traffic.",
                                        type="object",
                                        required=["name"],
                                        properties={
                                            "kind": V1JSONSchemaProps(
                                                type="string",
                                                enum=["Service", "TraefikService"],
                                            ),
                                            "name": V1JSONSchemaProps(
                                                description="Name is a reference to a Kubernetes Service object (for a load-balancer of servers), or to a TraefikServic                               object (service load-balancer, mirroring, etc). The differentiation between the two is specified in the Kind field.",
                                                type="string",
                                            ),
                                            "namespace": V1JSONSchemaProps(
                                                type="string",
                                            ),
                                            "passHostHeader": V1JSONSchemaProps(
                                                type="boolean",
                                            ),
                                            "port": V1JSONSchemaProps(
                                                any_of=[
                                                    V1JSONSchemaProps(
                                                        type="integer",
                                                    ),
                                                    V1JSONSchemaProps(
                                                        type="string",
                                                    ),
                                                ],
                                                x_kubernetes_int_or_string=True,
                                            ),
                                            "responseForwarding": V1JSONSchemaProps(
                                                description="ResponseForwarding holds configuration for the forward of the response.",
                                                type="object",
                                                properties={
                                                    "flushInterval": V1JSONSchemaProps(
                                                        type="string",
                                                    )
                                                },
                                            ),
                                            "scheme": V1JSONSchemaProps(
                                                type="string",
                                            ),
                                            "serversTransport": V1JSONSchemaProps(
                                                type="string",
                                            ),
                                            "sticky": V1JSONSchemaProps(
                                                description="Sticky holds the sticky configuration.",
                                                type="object",
                                                properties={
                                                    "cookie": V1JSONSchemaProps(
                                                        description="Cookie holds the sticky configuration based on cookie",
                                                        type="object",
                                                        properties={
                                                            "httpOnly": V1JSONSchemaProps(
                                                                type="boolean",
                                                            ),
                                                            "name": V1JSONSchemaProps(
                                                                type="string",
                                                            ),
                                                            "sameSite": V1JSONSchemaProps(
                                                                type="string",
                                                            ),
                                                            "secure": V1JSONSchemaProps(
                                                                type="boolean",
                                                            ),
                                                        },
                                                    )
                                                },
                                            ),
                                            "strategy": V1JSONSchemaProps(
                                                type="string",
                                            ),
                                            "weight": V1JSONSchemaProps(
                                                description="Weight should only be specified when Name references a TraefikService object (and to be precise, one that embeds a Weighted Round Robin).",
                                                type="integer",
                                            ),
                                        },
                                    ),
                                    "status": V1JSONSchemaProps(
                                        type="array",
                                        items={
                                            "type": "string",
                                        },
                                    ),
                                },
                            ),
                            "forwardAuth": V1JSONSchemaProps(
                                description="ForwardAuth holds the http forward authentication configuration.",
                                type="object",
                                properties={
                                    "address": V1JSONSchemaProps(
                                        type="string",
                                    ),
                                    "authRequestHeaders": V1JSONSchemaProps(
                                        type="array",
                                        items={
                                            "type": "string",
                                        },
                                    ),
                                    "authResponseHeaders": V1JSONSchemaProps(
                                        type="array",
                                        items={
                                            "type": "string",
                                        },
                                    ),
                                    "authResponseHeadersRegex": V1JSONSchemaProps(
                                        type="string",
                                    ),
                                    "tls": V1JSONSchemaProps(
                                        description="ClientTLS holds TLS specific configurations as client.",
                                        type="object",
                                        properties={
                                            "caOptional": V1JSONSchemaProps(
                                                type="string",
                                            ),
                                            "caSecret": V1JSONSchemaProps(
                                                type="string",
                                            ),
                                            "certSecret": V1JSONSchemaProps(
                                                type="string",
                                            ),
                                            "insecureSkipVerify": V1JSONSchemaProps(
                                                type="boolean",
                                            ),
                                        },
                                    ),
                                    "trustForwardHeader": V1JSONSchemaProps(
                                        type="boolean",
                                    ),
                                },
                            ),
                            "headers": V1JSONSchemaProps(
                                description="Headers holds the custom header configuration.",
                                type="object",
                                properties={
                                    "accessControlAllowCredentials": V1JSONSchemaProps(
                                        description="AccessControlAllowCredentials is only valid if true. false is ignored.",
                                        type="boolean",
                                    ),
                                    "accessControlAllowHeaders": V1JSONSchemaProps(
                                        description="AccessControlAllowHeaders must be used in response to a preflight request with Access-Control-Request-Headers set.",
                                        type="array",
                                        items={
                                            "type": "string",
                                        },
                                    ),
                                    "accessControlAllowMethods": V1JSONSchemaProps(
                                        type="array",
                                        items={
                                            "type": "string",
                                        },
                                    ),
                                    "accessControlAllowOriginList": V1JSONSchemaProps(
                                        type="array",
                                        items={
                                            "type": "string",
                                        },
                                    ),
                                    "accessControlAllowOriginListRegex": V1JSONSchemaProps(
                                        type="array",
                                        items={
                                            "type": "string",
                                        },
                                    ),
                                    "accessControlExposeHeaders": V1JSONSchemaProps(
                                        type="array",
                                        items={
                                            "type": "string",
                                        },
                                    ),
                                    "accessControlMaxAge": V1JSONSchemaProps(
                                        type="integer",
                                        format="int64",
                                    ),
                                    "addVaryHeader": V1JSONSchemaProps(
                                        type="boolean",
                                    ),
                                    "allowedHosts": V1JSONSchemaProps(
                                        type="array",
                                        items={
                                            "type": "string",
                                        },
                                    ),
                                    "browserXssFilter": V1JSONSchemaProps(
                                        type="boolean",
                                    ),
                                    "contentSecurityPolicy": V1JSONSchemaProps(
                                        type="string",
                                    ),
                                    "contentTypeNosniff": V1JSONSchemaProps(
                                        type="boolean",
                                    ),
                                    "customBrowserXSSValue": V1JSONSchemaProps(
                                        type="string",
                                    ),
                                    "customFrameOptionsValue": V1JSONSchemaProps(
                                        type="string",
                                    ),
                                    "customRequestHeaders": V1JSONSchemaProps(
                                        type="object",
                                        additional_properties={
                                            "type": "string",
                                        },
                                    ),
                                    "featurePolicy": V1JSONSchemaProps(
                                        description="Deprecated: use PermissionsPolicy instead.",
                                        type="string",
                                    ),
                                    "forceSTSHeader": V1JSONSchemaProps(
                                        type="boolean",
                                    ),
                                    "frameDeny": V1JSONSchemaProps(
                                        type="boolean",
                                    ),
                                    "hostsProxyHeaders": V1JSONSchemaProps(
                                        type="array",
                                        items={
                                            "type": "string",
                                        },
                                    ),
                                    "isDevelopment": V1JSONSchemaProps(
                                        type="boolean",
                                    ),
                                    "permissionsPolicy": V1JSONSchemaProps(
                                        type="string",
                                    ),
                                    "publicKey": V1JSONSchemaProps(
                                        type="string",
                                    ),
                                    "referrerPolicy": V1JSONSchemaProps(
                                        type="string",
                                    ),
                                    "sslForceHost": V1JSONSchemaProps(
                                        description="Deprecated: use RedirectRegex instead.",
                                        type="boolean",
                                    ),
                                    "sslHost": V1JSONSchemaProps(
                                        description="Deprecated: use RedirectRegex instead.",
                                        type="string",
                                    ),
                                    "sslProxyHeaders": V1JSONSchemaProps(
                                        type="object",
                                        additional_properties={
                                            "type": "string",
                                        },
                                    ),
                                    "sslRedirect": V1JSONSchemaProps(
                                        type="boolean",
                                    ),
                                    "sslTemporaryRedirect": V1JSONSchemaProps(
                                        type="boolean",
                                    ),
                                    "stsIncludeSubdomains": V1JSONSchemaProps(
                                        type="boolean",
                                    ),
                                    "stsPreload": V1JSONSchemaProps(
                                        type="boolean",
                                    ),
                                    "stsSeconds": V1JSONSchemaProps(
                                        type="integer",
                                        format="int64",
                                    ),
                                },
                            ),
                            "inFlightReq": V1JSONSchemaProps(
                                description="InFlightReq limits the number of requests being processed and served concurrently.",
                                type="object",
                                properties={
                                    "amount": V1JSONSchemaProps(
                                        type="integer",
                                        format="int64",
                                    ),
                                    "sourceCriterion": V1JSONSchemaProps(
                                        description="SourceCriterion defines what criterion is used to group requests as originating from a common source. If none are set, the default is to use the request's remote address field. All fields are mutually exclusive.",
                                        type="object",
                                        properties={
                                            "ipStrategy": V1JSONSchemaProps(
                                                description="IPStrategy holds the ip strategy configuration.",
                                                type="object",
                                                properties={
                                                    "depth": V1JSONSchemaProps(
                                                        type="integer",
                                                    ),
                                                    "excludedIPs": V1JSONSchemaProps(
                                                        type="array",
                                                        items={
                                                            "type": "string",
                                                        },
                                                    ),
                                                },
                                            ),
                                            "requestHeaderName": V1JSONSchemaProps(
                                                type="string",
                                            ),
                                            "requestHost": V1JSONSchemaProps(
                                                type="boolean",
                                            ),
                                        },
                                    ),
                                },
                            ),
                            "ipWhiteList": V1JSONSchemaProps(
                                description="IPWhiteList holds the ip white list configuration.",
                                type="object",
                                properties={
                                    "ipStrategy": V1JSONSchemaProps(
                                        description="IPStrategy holds the ip strategy configuration.",
                                        type="object",
                                        properties={
                                            "depth": V1JSONSchemaProps(
                                                type="integer",
                                            ),
                                            "excludedIPs": V1JSONSchemaProps(
                                                type="array",
                                                items={
                                                    "type": "string",
                                                },
                                            ),
                                        },
                                    ),
                                    "sourceRange": V1JSONSchemaProps(
                                        type="array",
                                        items={
                                            "type": "string",
                                        },
                                    ),
                                },
                            ),
                            "passTLSClientCert": V1JSONSchemaProps(
                                description="PassTLSClientCert holds the TLS client cert headers configuration.",
                                type="object",
                                properties={
                                    "info": V1JSONSchemaProps(
                                        description="TLSClientCertificateInfo holds the client TLS certificate info configuration.",
                                        type="object",
                                        properties={
                                            "issuer": V1JSONSchemaProps(
                                                description="TLSClientCertificateIssuerDNInfo holds the client TLS certificate distinguished name info configuration. cf https://tools.ietf.org/html/rfc3739",
                                                type="object",
                                                properties={
                                                    "commonName": V1JSONSchemaProps(
                                                        type="boolean",
                                                    ),
                                                    "country": V1JSONSchemaProps(
                                                        type="boolean",
                                                    ),
                                                    "domainComponent": V1JSONSchemaProps(
                                                        type="boolean",
                                                    ),
                                                    "organization": V1JSONSchemaProps(
                                                        type="boolean",
                                                    ),
                                                    "province": V1JSONSchemaProps(
                                                        type="boolean",
                                                    ),
                                                    "serialNumber": V1JSONSchemaProps(
                                                        type="boolean",
                                                    ),
                                                },
                                            ),
                                            "notAfter": V1JSONSchemaProps(
                                                type="boolean",
                                            ),
                                            "notBefore": V1JSONSchemaProps(
                                                type="boolean",
                                            ),
                                            "sans": V1JSONSchemaProps(
                                                type="boolean",
                                            ),
                                            "serialNumber": V1JSONSchemaProps(
                                                type="boolean",
                                            ),
                                            "subject": V1JSONSchemaProps(
                                                description="TLSClientCertificateSubjectDNInfo holds the client TLS certificate distinguished name info configuration. cf https://tools.ietf.org/html/rfc3739",
                                                type="object",
                                                properties={
                                                    "commonName": V1JSONSchemaProps(
                                                        type="boolean",
                                                    ),
                                                    "country": V1JSONSchemaProps(
                                                        type="boolean",
                                                    ),
                                                    "domainComponent": V1JSONSchemaProps(
                                                        type="boolean",
                                                    ),
                                                    "locality": V1JSONSchemaProps(
                                                        type="boolean",
                                                    ),
                                                    "organization": V1JSONSchemaProps(
                                                        type="boolean",
                                                    ),
                                                    "organizationalUnit": V1JSONSchemaProps(
                                                        type="boolean",
                                                    ),
                                                    "province": V1JSONSchemaProps(
                                                        type="boolean",
                                                    ),
                                                    "serialNumber": V1JSONSchemaProps(
                                                        type="boolean",
                                                    ),
                                                },
                                            ),
                                        },
                                    ),
                                    "pem": V1JSONSchemaProps(
                                        type="boolean",
                                    ),
                                },
                            ),
                            "plugin": V1JSONSchemaProps(
                                type="object",
                                additional_properties={
                                    "x-kubernetes-preserve-unknown-fields": True
                                },
                            ),
                            "rateLimit": V1JSONSchemaProps(
                                description="RateLimit holds the rate limiting configuration for a given router.",
                                type="object",
                                properties={
                                    "average": V1JSONSchemaProps(
                                        type="integer",
                                        format="int64",
                                    ),
                                    "burst": V1JSONSchemaProps(
                                        type="integer",
                                        format="int64",
                                    ),
                                    "period": V1JSONSchemaProps(
                                        any_of=[
                                            V1JSONSchemaProps(
                                                type="integer",
                                            ),
                                            V1JSONSchemaProps(
                                                type="string",
                                            ),
                                        ],
                                        x_kubernetes_int_or_string=True,
                                    ),
                                    "sourceCriterion": V1JSONSchemaProps(
                                        description="SourceCriterion defines what criterion is used to group requests as originating from a common source. If none are set, the default is to use the request's remote address field. All fields are mutually exclusive.",
                                        type="object",
                                        properties={
                                            "ipStrategy": V1JSONSchemaProps(
                                                description="IPStrategy holds the ip strategy configuration.",
                                                type="object",
                                                properties={
                                                    "depth": V1JSONSchemaProps(
                                                        type="integer",
                                                    ),
                                                    "excludedIPs": V1JSONSchemaProps(
                                                        type="array",
                                                        items={
                                                            "type": "string",
                                                        },
                                                    ),
                                                },
                                            ),
                                            "requestHeaderName": V1JSONSchemaProps(
                                                type="string",
                                            ),
                                            "requestHost": V1JSONSchemaProps(
                                                type="boolean",
                                            ),
                                        },
                                    ),
                                },
                            ),
                            "redirectRegex": V1JSONSchemaProps(
                                description="RedirectRegex holds the redirection configuration.",
                                type="object",
                                properties={
                                    "permanent": V1JSONSchemaProps(
                                        type="boolean",
                                    ),
                                    "regex": V1JSONSchemaProps(
                                        type="string",
                                    ),
                                    "replacement": V1JSONSchemaProps(
                                        type="string",
                                    ),
                                },
                            ),
                            "redirectScheme": V1JSONSchemaProps(
                                description="RedirectScheme holds the scheme redirection configuration.",
                                type="object",
                                properties={
                                    "permanent": V1JSONSchemaProps(
                                        type="boolean",
                                    ),
                                    "port": V1JSONSchemaProps(
                                        type="string",
                                    ),
                                    "scheme": V1JSONSchemaProps(
                                        type="string",
                                    ),
                                },
                            ),
                            "replacePath": V1JSONSchemaProps(
                                description="ReplacePath holds the ReplacePath configuration.",
                                type="object",
                                properties={
                                    "path": V1JSONSchemaProps(
                                        type="string",
                                    ),
                                },
                            ),
                            "replacePathRegex": V1JSONSchemaProps(
                                description="ReplacePathRegex holds the ReplacePathRegex configuration.",
                                type="object",
                                properties={
                                    "regex": V1JSONSchemaProps(
                                        type="string",
                                    ),
                                    "replacement": V1JSONSchemaProps(
                                        type="string",
                                    ),
                                },
                            ),
                            "retry": V1JSONSchemaProps(
                                description="Retry holds the retry configuration.",
                                type="object",
                                properties={
                                    "attempts": V1JSONSchemaProps(
                                        type="integer",
                                    ),
                                    "initialInterval": V1JSONSchemaProps(
                                        any_of=[
                                            V1JSONSchemaProps(
                                                type="integer",
                                            ),
                                            V1JSONSchemaProps(
                                                type="string",
                                            ),
                                        ],
                                        x_kubernetes_int_or_string=True,
                                    ),
                                },
                            ),
                            "stripPrefix": V1JSONSchemaProps(
                                description="StripPrefix holds the StripPrefix configuration.",
                                type="object",
                                properties={
                                    "forceSlash": V1JSONSchemaProps(
                                        type="boolean",
                                    ),
                                    "prefixes": V1JSONSchemaProps(
                                        type="array",
                                        items={
                                            "type": "string",
                                        },
                                    ),
                                },
                            ),
                            "stripPrefixRegex": V1JSONSchemaProps(
                                description="StripPrefixRegex holds the StripPrefixRegex configuration.",
                                type="object",
                                properties={
                                    "regex": V1JSONSchemaProps(
                                        type="array",
                                        items={
                                            "type": "string",
                                        },
                                    ),
                                },
                            ),
                        },
                    ),
                },
            ),
        )
    ],
)

middlewaretcp_crd = CreateCustomResourceDefinition(
    crd_name="middlewaretcps.traefik.containo.us",
    app_name=traefik_name,
    group="traefik.containo.us",
    names=CustomResourceDefinitionNames(
        kind="MiddlewareTCP",
        list_kind="MiddlewareTCPList",
        plural="middlewaretcps",
        singular="middlewaretcp",
    ),
    annotations={
        "controller-gen.kubebuilder.io/version": "v0.6.2",
    },
    versions=[
        CustomResourceDefinitionVersion(
            name="v1alpha1",
            open_apiv3_schema=V1JSONSchemaProps(
                description="MiddlewareTCP is a specification for a MiddlewareTCP resource.",
                type="object",
                required=["metadata", "spec"],
                properties={
                    "apiVersion": V1JSONSchemaProps(
                        description="APIVersion defines the versioned schema of this representation of an object. Servers should convert recognized schemas to the latest internal value, and may reject unrecognized values. More info: https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#resources",
                        type="string",
                    ),
                    "kind": V1JSONSchemaProps(
                        description="Kind is a string value representing the REST resource this object represents. Servers may infer this from the endpoint the client submits requests to. Cannot be updated. In CamelCase. More info: https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#types-kinds",
                        type="string",
                    ),
                    "metadata": V1JSONSchemaProps(type="object"),
                    "spec": V1JSONSchemaProps(
                        description="MiddlewareTCPSpec holds the MiddlewareTCP configuration.",
                        type="object",
                        properties={
                            "inFlightConn": V1JSONSchemaProps(
                                description="TCPInFlightConn holds the TCP in flight connection configuration.",
                                type="object",
                                properties={
                                    "amount": V1JSONSchemaProps(
                                        type="integer",
                                        format="int64",
                                    ),
                                },
                            ),
                            "ipWhiteList": V1JSONSchemaProps(
                                description="TCPIPWhiteList holds the TCP ip white list configuration.",
                                type="object",
                                properties={
                                    "sourceRange": V1JSONSchemaProps(
                                        type="array",
                                        items={
                                            "type": "string",
                                        },
                                    ),
                                },
                            ),
                        },
                    ),
                },
            ),
        )
    ],
)

serverstransport_crd = CreateCustomResourceDefinition(
    crd_name="serverstransports.traefik.containo.us",
    app_name=traefik_name,
    group="traefik.containo.us",
    names=CustomResourceDefinitionNames(
        kind="ServersTransport",
        list_kind="ServersTransportList",
        plural="serverstransports",
        singular="serverstransport",
    ),
    annotations={
        "controller-gen.kubebuilder.io/version": "v0.6.2",
    },
    versions=[
        CustomResourceDefinitionVersion(
            name="v1alpha1",
            open_apiv3_schema=V1JSONSchemaProps(
                description="ServersTransport is a specification for a ServersTransport resource.",
                type="object",
                required=["metadata", "spec"],
                properties={
                    "apiVersion": V1JSONSchemaProps(
                        description="APIVersion defines the versioned schema of this representation of an object. Servers should convert recognized schemas to the latest internal value, and may reject unrecognized values. More info: https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#resources",
                        type="string",
                    ),
                    "kind": V1JSONSchemaProps(
                        description="Kind is a string value representing the REST resource this object represents. Servers may infer this from the endpoint the client submits requests to. Cannot be updated. In CamelCase. More info: https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#types-kinds",
                        type="string",
                    ),
                    "metadata": V1JSONSchemaProps(type="object"),
                    "spec": V1JSONSchemaProps(
                        description="ServersTransportSpec options to configure communication between Traefik and the servers.",
                        type="object",
                        properties={
                            "certificatesSecrets": V1JSONSchemaProps(
                                description="Certificates for mTLS.",
                                type="array",
                                items={
                                    "type": "string",
                                },
                            ),
                            "disableHTTP2": V1JSONSchemaProps(
                                description="Disable HTTP/2 for connections with backend servers.",
                                type="boolean",
                            ),
                            "forwardingTimeouts": V1JSONSchemaProps(
                                description="Timeouts for requests forwarded to the backend servers.",
                                type="object",
                                properties={
                                    "dialTimeout": V1JSONSchemaProps(
                                        description="DialTimeout is the amount of time to wait until a connection to a backend server can be established. If zero, no timeout exists.",
                                        any_of=[
                                            V1JSONSchemaProps(
                                                type="integer",
                                            ),
                                            V1JSONSchemaProps(
                                                type="string",
                                            ),
                                        ],
                                        x_kubernetes_int_or_string=True,
                                    ),
                                    "idleConnTimeout": V1JSONSchemaProps(
                                        description="IdleConnTimeout is the maximum period for which an idle HTTP keep-alive connection will remain open before closing itself.",
                                        any_of=[
                                            V1JSONSchemaProps(
                                                type="integer",
                                            ),
                                            V1JSONSchemaProps(
                                                type="string",
                                            ),
                                        ],
                                        x_kubernetes_int_or_string=True,
                                    ),
                                    "pingTimeout": V1JSONSchemaProps(
                                        any_of=[
                                            V1JSONSchemaProps(
                                                type="integer",
                                            ),
                                            V1JSONSchemaProps(
                                                type="string",
                                            ),
                                        ],
                                        x_kubernetes_int_or_string=True,
                                    ),
                                    "readIdleTimeout": V1JSONSchemaProps(
                                        any_of=[
                                            V1JSONSchemaProps(
                                                type="integer",
                                            ),
                                            V1JSONSchemaProps(
                                                type="string",
                                            ),
                                        ],
                                        x_kubernetes_int_or_string=True,
                                    ),
                                    "responseHeaderTimeout": V1JSONSchemaProps(
                                        any_of=[
                                            V1JSONSchemaProps(
                                                type="integer",
                                            ),
                                            V1JSONSchemaProps(
                                                type="string",
                                            ),
                                        ],
                                        x_kubernetes_int_or_string=True,
                                    ),
                                },
                            ),
                            "insecureSkipVerify": V1JSONSchemaProps(
                                description="Disable SSL certificate verification.",
                                type="boolean",
                            ),
                            "maxIdleConnsPerHost": V1JSONSchemaProps(
                                description="If non-zero, controls the maximum idle (keep-alive) to keep per-host. If zero, DefaultMaxIdleConnsPerHost is used.",
                                type="integer",
                            ),
                            "peerCertURI": V1JSONSchemaProps(
                                description="URI used to match against SAN URI during the peer certificate verification.",
                                type="string",
                            ),
                            "rootCAsSecrets": V1JSONSchemaProps(
                                description="Add cert file for self-signed certificate.",
                                type="array",
                                items={
                                    "type": "string",
                                },
                            ),
                            "serverName": V1JSONSchemaProps(
                                description="ServerName used to contact the server.",
                                type="string",
                            ),
                        },
                    ),
                },
            ),
        )
    ],
)

tlsoption_crd = CreateCustomResourceDefinition(
    crd_name="tlsoptions.traefik.containo.us",
    app_name=traefik_name,
    group="traefik.containo.us",
    names=CustomResourceDefinitionNames(
        kind="TLSOption",
        list_kind="TLSOptionList",
        plural="tlsoptions",
        singular="tlsoption",
    ),
    annotations={
        "controller-gen.kubebuilder.io/version": "v0.6.2",
    },
    versions=[
        CustomResourceDefinitionVersion(
            name="v1alpha1",
            open_apiv3_schema=V1JSONSchemaProps(
                description="TLSOption is a specification for a TLSOption resource.",
                type="object",
                required=["metadata", "spec"],
                properties={
                    "apiVersion": V1JSONSchemaProps(
                        description="APIVersion defines the versioned schema of this representation of an object. Servers should convert recognized schemas to the latest internal value, and may reject unrecognized values. More info: https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#resources",
                        type="string",
                    ),
                    "kind": V1JSONSchemaProps(
                        description="Kind is a string value representing the REST resource this object represents. Servers may infer this from the endpoint the client submits requests to. Cannot be updated. In CamelCase. More info: https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#types-kinds",
                        type="string",
                    ),
                    "metadata": V1JSONSchemaProps(type="object"),
                    "spec": V1JSONSchemaProps(
                        description="TLSOptionSpec configures TLS for an entry point.",
                        type="object",
                        properties={
                            "alpnProtocols": V1JSONSchemaProps(
                                type="array",
                                items={
                                    "type": "string",
                                },
                            ),
                            "cipherSuites": V1JSONSchemaProps(
                                type="array",
                                items={
                                    "type": "string",
                                },
                            ),
                            "clientAuth": V1JSONSchemaProps(
                                description="ClientAuth defines the parameters of the client authentication part of the TLS connection, if any.",
                                type="object",
                                properties={
                                    "clientAuthType": V1JSONSchemaProps(
                                        description="ClientAuthType defines the client authentication type to apply.",
                                        enum=[
                                            "NoClientCert",
                                            "RequestClientCert",
                                            "RequireAnyClientCert",
                                            "VerifyClientCertIfGiven",
                                            "RequireAndVerifyClientCert",
                                        ],
                                        type="string",
                                    ),
                                    "secretNames": V1JSONSchemaProps(
                                        description="SecretName is the name of the referenced Kubernetes Secret to specify the certificate details.",
                                        type="array",
                                        items={
                                            "type": "string",
                                        },
                                    ),
                                },
                            ),
                            "curvePreferences": V1JSONSchemaProps(
                                type="array",
                                items={
                                    "type": "string",
                                },
                            ),
                            "maxVersion": V1JSONSchemaProps(
                                type="string",
                            ),
                            "minVersion": V1JSONSchemaProps(
                                type="string",
                            ),
                            "preferServerCipherSuites": V1JSONSchemaProps(
                                type="boolean",
                            ),
                            "sniStrict": V1JSONSchemaProps(
                                type="boolean",
                            ),
                        },
                    ),
                },
            ),
        )
    ],
)

tlsstore_crd = CreateCustomResourceDefinition(
    crd_name="tlsstores.traefik.containo.us",
    app_name=traefik_name,
    group="traefik.containo.us",
    names=CustomResourceDefinitionNames(
        kind="TLSStore",
        list_kind="TLSStoreList",
        plural="tlsstores",
        singular="tlsstore",
    ),
    annotations={
        "controller-gen.kubebuilder.io/version": "v0.6.2",
    },
    versions=[
        CustomResourceDefinitionVersion(
            name="v1alpha1",
            open_apiv3_schema=V1JSONSchemaProps(
                description="TLSStore is a specification for a TLSStore resource.",
                type="object",
                required=["metadata", "spec"],
                properties={
                    "apiVersion": V1JSONSchemaProps(
                        description="APIVersion defines the versioned schema of this representation of an object. Servers should convert recognized schemas to the latest internal value, and may reject unrecognized values. More info: https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#resources",
                        type="string",
                    ),
                    "kind": V1JSONSchemaProps(
                        description="Kind is a string value representing the REST resource this object represents. Servers may infer this from the endpoint the client submits requests to. Cannot be updated. In CamelCase. More info: https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#types-kinds",
                        type="string",
                    ),
                    "metadata": V1JSONSchemaProps(type="object"),
                    "spec": V1JSONSchemaProps(
                        description="TLSStoreSpec configures a TLSStore resource.",
                        type="object",
                        properties={
                            "defaultCertificate": V1JSONSchemaProps(
                                description="DefaultCertificate holds a secret name for the TLSOption resource.",
                                required=["secretName"],
                                type="object",
                                properties={
                                    "secretName": V1JSONSchemaProps(
                                        description="SecretName is the name of the referenced Kubernetes Secret to specify the certificate details.",
                                        type="string",
                                    ),
                                },
                            )
                        },
                    ),
                },
            ),
        )
    ],
)

traefikservice_crd = CreateCustomResourceDefinition(
    crd_name="traefikservices.traefik.containo.us",
    app_name=traefik_name,
    group="traefik.containo.us",
    names=CustomResourceDefinitionNames(
        kind="TraefikService",
        list_kind="TraefikServiceList",
        plural="traefikservices",
        singular="traefikservice",
    ),
    annotations={
        "controller-gen.kubebuilder.io/version": "v0.6.2",
    },
    versions=[
        CustomResourceDefinitionVersion(
            name="v1alpha1",
            open_apiv3_schema=V1JSONSchemaProps(
                description="TraefikService is the specification for a service (that an IngressRoute refers to) that is usually not a terminal service (i.e. not a pod of servers), as opposed to a Kubernetes Service. That is to say, it usually refers to other (children) services, which themselves can be TraefikServices or Services.",
                type="object",
                required=["metadata", "spec"],
                properties={
                    "apiVersion": V1JSONSchemaProps(
                        description="APIVersion defines the versioned schema of this representation of an object. Servers should convert recognized schemas to the latest internal value, and may reject unrecognized values. More info: https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#resources",
                        type="string",
                    ),
                    "kind": V1JSONSchemaProps(
                        description="Kind is a string value representing the REST resource this object represents. Servers may infer this from the endpoint the client submits requests to. Cannot be updated. In CamelCase. More info: https://git.k8s.io/community/contributors/devel/sig-architecture/api-conventions.md#types-kinds",
                        type="string",
                    ),
                    "metadata": V1JSONSchemaProps(type="object"),
                    "spec": V1JSONSchemaProps(
                        description="ServiceSpec defines whether a TraefikService is a load-balancer of services or a mirroring service.",
                        type="object",
                        properties={
                            "mirroring": V1JSONSchemaProps(
                                description="Mirroring defines a mirroring service, which is composed of a main load-balancer, and a list of mirrors.",
                                type="object",
                                required=["name"],
                                properties={
                                    "kind": V1JSONSchemaProps(
                                        type="string",
                                        enum=["Service", "TraefikService"],
                                    ),
                                    "maxBodySize": V1JSONSchemaProps(
                                        type="integer",
                                        format="int64",
                                    ),
                                    "mirrors": V1JSONSchemaProps(
                                        type="array",
                                        items={
                                            "description": "MirrorService defines one of the mirrors of a Mirroring service.",
                                            "type": "object",
                                            "required": ["name"],
                                            "properties": {
                                                "kind": V1JSONSchemaProps(
                                                    type="string",
                                                    enum=["Service", "TraefikService"],
                                                ),
                                                "name": V1JSONSchemaProps(
                                                    description="Name is a reference to a Kubernetes Service object (for a load-balancer of servers), or to a TraefikServic                               object (service load-balancer, mirroring, etc). The differentiation between the two is specified in the Kind field.",
                                                    type="string",
                                                ),
                                                "passHostHeader": V1JSONSchemaProps(
                                                    type="boolean",
                                                ),
                                                "percent": V1JSONSchemaProps(
                                                    type="integer",
                                                ),
                                                "port": V1JSONSchemaProps(
                                                    any_of=[
                                                        V1JSONSchemaProps(
                                                            type="integer",
                                                        ),
                                                        V1JSONSchemaProps(
                                                            type="string",
                                                        ),
                                                    ],
                                                    x_kubernetes_int_or_string=True,
                                                ),
                                                "responseForwarding": V1JSONSchemaProps(
                                                    description="ResponseForwarding holds configuration for the forward of the response.",
                                                    type="object",
                                                    properties={
                                                        "flushInterval": V1JSONSchemaProps(
                                                            type="string",
                                                        )
                                                    },
                                                ),
                                                "scheme": V1JSONSchemaProps(
                                                    type="string",
                                                ),
                                                "serversTransport": V1JSONSchemaProps(
                                                    type="string",
                                                ),
                                                "sticky": V1JSONSchemaProps(
                                                    description="Sticky holds the sticky configuration.",
                                                    type="object",
                                                    properties={
                                                        "cookie": V1JSONSchemaProps(
                                                            description="Cookie holds the sticky configuration based on cookie",
                                                            type="object",
                                                            properties={
                                                                "httpOnly": V1JSONSchemaProps(
                                                                    type="boolean",
                                                                ),
                                                                "name": V1JSONSchemaProps(
                                                                    type="string",
                                                                ),
                                                                "sameSite": V1JSONSchemaProps(
                                                                    type="string",
                                                                ),
                                                                "secure": V1JSONSchemaProps(
                                                                    type="boolean",
                                                                ),
                                                            },
                                                        )
                                                    },
                                                ),
                                                "strategy": V1JSONSchemaProps(
                                                    type="string",
                                                ),
                                                "weight": V1JSONSchemaProps(
                                                    description="Weight should only be specified when Name references a TraefikService object (and to be precise, one that embeds a Weighted Round Robin).",
                                                    type="integer",
                                                ),
                                            },
                                        },
                                    ),
                                    "name": V1JSONSchemaProps(
                                        description="Name is a reference to a Kubernetes Service object (for a load-balancer of servers), or to a TraefikService object (service load-balancer, mirroring, etc). The differentiation between the two is specified in the Kind field.",
                                        type="string",
                                    ),
                                    "namespace": V1JSONSchemaProps(
                                        type="string",
                                    ),
                                    "passHostHeader": V1JSONSchemaProps(
                                        type="boolean",
                                    ),
                                    "port": V1JSONSchemaProps(
                                        any_of=[
                                            V1JSONSchemaProps(
                                                type="integer",
                                            ),
                                            V1JSONSchemaProps(
                                                type="string",
                                            ),
                                        ],
                                        x_kubernetes_int_or_string=True,
                                    ),
                                    "responseForwarding": V1JSONSchemaProps(
                                        description="ResponseForwarding holds configuration for the forward of the response.",
                                        type="object",
                                        properties={
                                            "flushInterval": V1JSONSchemaProps(
                                                type="string",
                                            )
                                        },
                                    ),
                                    "scheme": V1JSONSchemaProps(
                                        type="string",
                                    ),
                                    "serversTransport": V1JSONSchemaProps(
                                        type="string",
                                    ),
                                    "sticky": V1JSONSchemaProps(
                                        description="Sticky holds the sticky configuration.",
                                        type="object",
                                        properties={
                                            "cookie": V1JSONSchemaProps(
                                                description="Cookie holds the sticky configuration based on cookie",
                                                type="object",
                                                properties={
                                                    "httpOnly": V1JSONSchemaProps(
                                                        type="boolean",
                                                    ),
                                                    "name": V1JSONSchemaProps(
                                                        type="string",
                                                    ),
                                                    "sameSite": V1JSONSchemaProps(
                                                        type="string",
                                                    ),
                                                    "secure": V1JSONSchemaProps(
                                                        type="boolean",
                                                    ),
                                                },
                                            )
                                        },
                                    ),
                                    "strategy": V1JSONSchemaProps(
                                        type="string",
                                    ),
                                    "weight": V1JSONSchemaProps(
                                        description="Weight should only be specified when Name references a TraefikService object (and to be precise, one that embeds a Weighted Round Robin).",
                                        type="integer",
                                    ),
                                },
                            ),
                            "weighted": V1JSONSchemaProps(
                                description="WeightedRoundRobin defines a load-balancer of services.",
                                type="object",
                                properties={
                                    "services": V1JSONSchemaProps(
                                        type="array",
                                        items={
                                            "description": "Service defines an upstream to proxy traffic.",
                                            "type": "object",
                                            "required": ["name"],
                                            "properties": {
                                                "kind": V1JSONSchemaProps(
                                                    type="string",
                                                    enum=["Service", "TraefikService"],
                                                ),
                                                "name": V1JSONSchemaProps(
                                                    description="Name is a reference to a Kubernetes Service object (for a load-balancer of servers), or to a TraefikServic                               object (service load-balancer, mirroring, etc). The differentiation between the two is specified in the Kind field.",
                                                    type="string",
                                                ),
                                                "passHostHeader": V1JSONSchemaProps(
                                                    type="boolean",
                                                ),
                                                "port": V1JSONSchemaProps(
                                                    any_of=[
                                                        V1JSONSchemaProps(
                                                            type="integer",
                                                        ),
                                                        V1JSONSchemaProps(
                                                            type="string",
                                                        ),
                                                    ],
                                                    x_kubernetes_int_or_string=True,
                                                ),
                                                "responseForwarding": V1JSONSchemaProps(
                                                    description="ResponseForwarding holds configuration for the forward of the response.",
                                                    type="object",
                                                    properties={
                                                        "flushInterval": V1JSONSchemaProps(
                                                            type="string",
                                                        )
                                                    },
                                                ),
                                                "scheme": V1JSONSchemaProps(
                                                    type="string",
                                                ),
                                                "serversTransport": V1JSONSchemaProps(
                                                    type="string",
                                                ),
                                                "sticky": V1JSONSchemaProps(
                                                    description="Sticky holds the sticky configuration.",
                                                    type="object",
                                                    properties={
                                                        "cookie": V1JSONSchemaProps(
                                                            description="Cookie holds the sticky configuration based on cookie",
                                                            type="object",
                                                            properties={
                                                                "httpOnly": V1JSONSchemaProps(
                                                                    type="boolean",
                                                                ),
                                                                "name": V1JSONSchemaProps(
                                                                    type="string",
                                                                ),
                                                                "sameSite": V1JSONSchemaProps(
                                                                    type="string",
                                                                ),
                                                                "secure": V1JSONSchemaProps(
                                                                    type="boolean",
                                                                ),
                                                            },
                                                        )
                                                    },
                                                ),
                                                "strategy": V1JSONSchemaProps(
                                                    type="string",
                                                ),
                                                "weight": V1JSONSchemaProps(
                                                    description="Weight should only be specified when Name references a TraefikService object (and to be precise, one that embeds a Weighted Round Robin).",
                                                    type="integer",
                                                ),
                                            },
                                        },
                                    ),
                                    "sticky": V1JSONSchemaProps(
                                        description="Sticky holds the sticky configuration.",
                                        type="object",
                                        properties={
                                            "cookie": V1JSONSchemaProps(
                                                description="Cookie holds the sticky configuration based on cookie",
                                                type="object",
                                                properties={
                                                    "httpOnly": V1JSONSchemaProps(
                                                        type="boolean",
                                                    ),
                                                    "name": V1JSONSchemaProps(
                                                        type="string",
                                                    ),
                                                    "sameSite": V1JSONSchemaProps(
                                                        type="string",
                                                    ),
                                                    "secure": V1JSONSchemaProps(
                                                        type="boolean",
                                                    ),
                                                },
                                            )
                                        },
                                    ),
                                },
                            ),
                        },
                    ),
                },
            ),
        )
    ],
)



from typing import Optional, Any, Dict, List, Union, cast
from typing_extensions import Literal
from textwrap import dedent

from phidata.infra.aws.api_client import AwsApiClient
from phidata.infra.aws.resource.base import AwsResource
from phidata.infra.aws.resource.ec2.subnet import Subnet
from phidata.infra.aws.resource.eks.cluster import EksCluster
from phidata.infra.aws.resource.iam.role import IamRole
from phidata.utils.cli_console import print_info, print_error
from phidata.utils.log import logger


class EksNodeGroup(AwsResource):
    """
    An Amazon EKS managed node group is an Amazon EC2 Auto Scaling group and associated EC2
    instances that are managed by Amazon Web Services for an Amazon EKS cluster.

    An Auto Scaling group is a group of EC2 instances that are combined into one management unit.
    When you set up an auto-scaling group, you specify a scaling policy and AWS will apply that policy to make sure
    that a certain number of instances is automatically running in your group. If the number of instances drops below a
    certain value, or if the load increases (depending on the policy),
    then AWS will automatically spin up new instances for you.

    https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/eks.html
    https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/eks.html#EKS.Client.create_nodegroup
    """

    resource_type = "EksNodeGroup"
    service_name = "eks"

    # Name for the node group
    name: str
    # The cluster to create the EksNodeGroup in
    eks_cluster: EksCluster

    # The IAM role to associate with your node group.
    # The Amazon EKS worker node kubelet daemon makes calls to Amazon Web Services APIs on your behalf.
    # Nodes receive permissions for these API calls through an IAM instance profile and associated policies.
    # Before you can launch nodes and register them into a cluster,
    # you must create an IAM role for those nodes to use when they are launched.

    # ARN for the node group IAM role to use
    node_role_arn: Optional[str] = None
    # If node_role_arn is None, a default role is created if create_role is True
    create_role: bool = True
    # If node_role is None, a default node_role is created using node_role_name
    node_role: Optional[IamRole] = None
    # Name for the default node_role when role is None, use "name-iam-role" if not provided
    node_role_name: Optional[str] = None
    # Provide a list of policy ARNs to attach to the node group role
    add_policy_arns: Optional[List[str]] = None

    # The scaling configuration details for the Auto Scaling group
    # Users can provide a dict for scaling config or use min/max/desired values below
    scaling_config: Optional[Dict[str, Union[str, int]]] = None
    # The minimum number of nodes that the managed node group can scale in to.
    min_size: Optional[int] = None
    # The maximum number of nodes that the managed node group can scale out to.
    max_size: Optional[int] = None
    # The current number of nodes that the managed node group should maintain.
    # WARNING: If you use Cluster Autoscaler, you shouldn't change the desired_size value directly,
    # as this can cause the Cluster Autoscaler to suddenly scale up or scale down.
    # Whenever this parameter changes, the number of worker nodes in the node group is updated to
    # the specified size. If this parameter is given a value that is smaller than the current number of
    # running worker nodes, the necessary number of worker nodes are terminated to match the given value.
    desired_size: Optional[int] = None
    # The root device disk size (in GiB) for your node group instances.
    # The default disk size is 20 GiB. If you specify launchTemplate,
    # then don't specify diskSize, or the node group deployment will fail.
    disk_size: Optional[int] = None
    # The subnets to use for the Auto Scaling group that is created for your node group.
    # If you specify launchTemplate, then don't specify SubnetId in your launch template,
    # or the node group deployment will fail.
    # For more information about using launch templates with Amazon EKS,
    # see Launch template support in the Amazon EKS User Guide.
    subnets: Optional[List[str]] = None
    # Filter subnets using availability zones
    subnet_az: Optional[Union[str, List[str]]] = None
    # Specify the instance types for a node group.
    # If you specify a GPU instance type, be sure to specify AL2_x86_64_GPU with the amiType parameter.
    # If you specify launchTemplate , then you can specify zero or one instance type in your launch template
    # or you can specify 0-20 instance types for instanceTypes .
    # If however, you specify an instance type in your launch template and specify any instanceTypes ,
    # the node group deployment will fail. If you don't specify an instance type in a launch template
    # or for instance_types, then t3.medium is used, by default. If you specify Spot for capacityType,
    # then we recommend specifying multiple values for instanceTypes .
    instance_types: Optional[List[str]] = None
    # The AMI type for your node group. GPU instance types should use the AL2_x86_64_GPU AMI type.
    # Non-GPU instances should use the AL2_x86_64 AMI type.
    # Arm instances should use the AL2_ARM_64 AMI type.
    # All types use the Amazon EKS optimized Amazon Linux 2 AMI.
    # If you specify launchTemplate , and your launch template uses a custom AMI,
    # then don't specify amiType , or the node group deployment will fail.
    ami_type: Optional[
        Literal[
            "AL2_x86_64",
            "AL2_x86_64_GPU",
            "AL2_ARM_64",
            "CUSTOM",
            "BOTTLEROCKET_ARM_64",
            "BOTTLEROCKET_x86_64",
        ]
    ] = None
    # The remote access (SSH) configuration to use with your node group.
    # If you specify launchTemplate, then don't specify remoteAccess, or the node group deployment will fail. For
    # Keys:
    #   ec2SshKey (string) -- The Amazon EC2 SSH key that provides access for SSH communication with the nodes
    #   in the managed node group. For more information, see Amazon EC2 key pairs and Linux instances in the
    #   Amazon Elastic Compute Cloud User Guide for Linux Instances .
    #   sourceSecurityGroups (list) -- The security groups that are allowed SSH access (port 22) to the nodes.
    #   If you specify an Amazon EC2 SSH key but do not specify a source security group when you create
    #   a managed node group, then port 22 on the nodes is opened to the internet (0.0.0.0/0).
    #   For more information, see Security Groups for Your VPC in the Amazon Virtual Private Cloud User Guide .
    remote_access: Optional[Dict[str, str]] = None
    # The Kubernetes labels to be applied to the nodes in the node group when they are created.
    labels: Optional[Dict[str, str]] = None
    # The Kubernetes taints to be applied to the nodes in the node group.
    taints: Optional[List[dict]] = None
    # The metadata to apply to the node group to assist with categorization and organization.
    # Each tag consists of a key and an optional value. You define both.
    # Node group tags do not propagate to any other resources associated with the node group,
    # such as the Amazon EC2 instances or subnets.
    tags: Optional[Dict[str, str]] = None
    # Unique, case-sensitive identifier that you provide to ensure the idempotency of the request.
    # This field is autopopulated if not provided.
    client_request_token: Optional[str] = None
    # An object representing a node group's launch template specification.
    # If specified, then do not specify instanceTypes, diskSize, or remoteAccess and make sure that the launch template
    # meets the requirements in launchTemplateSpecification .
    launch_template: Optional[Dict[str, str]] = None
    # The node group update configuration.
    update_config: Optional[Dict[str, int]] = None
    # The capacity type for your node group.
    capacity_type: Optional[Literal["ON_DEMAND", "SPOT"]] = None
    # The Kubernetes version to use for your managed nodes.
    # By default, the Kubernetes version of the cluster is used, and this is the only accepted specified value.
    # If you specify launchTemplate , and your launch template uses a custom AMI,
    # then don't specify version , or the node group deployment will fail.
    version: Optional[str] = None
    # The AMI version of the Amazon EKS optimized AMI to use with your node group.
    # By default, the latest available AMI version for the node group's current Kubernetes version is used.
    release_version: Optional[str] = None

    # provided by api on create
    created_at: Optional[str] = None
    nodegroup_status: Optional[str] = None

    # bump the wait time for Eks to 30 seconds
    waiter_delay = 30

    def _create(self, aws_client: AwsApiClient) -> bool:
        """Creates a NodeGroup for your Amazon EKS cluster.

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        print_info(f"Creating {self.get_resource_type()}: {self.get_resource_name()}")

        # Step 1: Get NodeGroup IamRole
        nodegroup_iam_role_arn = self.node_role_arn
        if nodegroup_iam_role_arn is None and self.create_role:
            # Create NodeGroup IamRole and get nodegroup_iam_role_arn
            nodegroup_iam_role = self.get_nodegroup_iam_role()
            try:
                nodegroup_iam_role.create(aws_client)
                nodegroup_iam_role_arn = nodegroup_iam_role.read(aws_client).arn
                print_info(
                    f"ARN for {nodegroup_iam_role.name}: {nodegroup_iam_role_arn}"
                )
            except Exception as e:
                print_error(
                    "NodeGroup IamRole creation failed, please fix and try again"
                )
                print_error(e)
                return False
        if nodegroup_iam_role_arn is None:
            print_error(f"IamRole ARN not available, please fix and try again")
            return False

        # Step 2: Get the subnets
        subnets: Optional[List[str]] = self.subnets
        if subnets is None:
            # Use subnets from EKSCluster if subnets not provided
            subnets = self.eks_cluster.get_subnets(aws_client=aws_client)
            # Filter subnets using availability zones
            if self.subnet_az is not None:
                azs_filter = []
                if isinstance(self.subnet_az, str):
                    azs_filter.append(self.subnet_az)
                elif isinstance(self.subnet_az, list):
                    azs_filter.extend(self.subnet_az)

                subnets = [
                    subnet_id
                    for subnet_id in subnets
                    if Subnet(id=subnet_id).get_availability_zone(aws_client=aws_client)
                    in azs_filter
                ]
            logger.debug(f"Using subnets from EKSCluster: {subnets}")
        # cast for type checker
        subnets = cast(List[str], subnets)

        # Step 3: Get the scaling_config
        scaling_config: Optional[Dict[str, Union[str, int]]] = self.scaling_config
        if scaling_config is None:
            # Build the scaling_config
            if self.min_size is not None:
                if scaling_config is None:
                    scaling_config = {}
                scaling_config["minSize"] = self.min_size
                # use min_size as the default for maxSize/desiredSize incase maxSize/desiredSize is not provided
                scaling_config["maxSize"] = self.min_size
                scaling_config["desiredSize"] = self.min_size
            if self.max_size is not None:
                if scaling_config is None:
                    scaling_config = {}
                scaling_config["maxSize"] = self.max_size
            if self.desired_size is not None:
                if scaling_config is None:
                    scaling_config = {}
                scaling_config["desiredSize"] = self.desired_size

        # create a dict of args which are not null, otherwise aws type validation fails
        not_null_args: Dict[str, Any] = {}
        if scaling_config is not None:
            not_null_args["scalingConfig"] = scaling_config
        if self.disk_size is not None:
            not_null_args["diskSize"] = self.disk_size
        if self.instance_types is not None:
            not_null_args["instanceTypes"] = self.instance_types
        if self.ami_type is not None:
            not_null_args["amiType"] = self.ami_type
        if self.remote_access is not None:
            not_null_args["remoteAccess"] = self.remote_access
        if self.labels is not None:
            not_null_args["labels"] = self.labels
        if self.taints is not None:
            not_null_args["taints"] = self.taints
        if self.tags is not None:
            not_null_args["tags"] = self.tags
        if self.client_request_token is not None:
            not_null_args["clientRequestToken"] = self.client_request_token
        if self.launch_template is not None:
            not_null_args["launchTemplate"] = self.launch_template
        if self.update_config is not None:
            not_null_args["updateConfig"] = self.update_config
        if self.capacity_type is not None:
            not_null_args["capacityType"] = self.capacity_type
        if self.version is not None:
            not_null_args["version"] = self.version
        if self.release_version is not None:
            not_null_args["release_version"] = self.release_version

        # Step 4: Create EksNodeGroup
        service_client = self.get_service_client(aws_client)
        try:
            create_response = service_client.create_nodegroup(
                clusterName=self.eks_cluster.name,
                nodegroupName=self.name,
                subnets=subnets,
                nodeRole=nodegroup_iam_role_arn,
                **not_null_args,
            )
            logger.debug(f"EksNodeGroup: {create_response}")
            nodegroup_dict = create_response.get("nodegroup", {})

            # Validate EksNodeGroup creation
            self.created_at = nodegroup_dict.get("createdAt", None)
            self.nodegroup_status = nodegroup_dict.get("status", None)
            logger.debug(f"created_at: {self.created_at}")
            logger.debug(f"nodegroup_status: {self.nodegroup_status}")
            if self.created_at is not None:
                print_info(f"EksNodeGroup created: {self.name}")
                self.active_resource = create_response
                return True
        except service_client.exceptions.ResourceInUseException:
            print_info(f"EksNodeGroup already exists: {self.name}")
            return True
        except Exception as e:
            print_error(f"{self.get_resource_type()} could not be created.")
            print_error(e)
        return False

    def post_create(self, aws_client: AwsApiClient) -> bool:

        # Wait for EksNodeGroup to be created
        if self.wait_for_creation:
            try:
                print_info(f"Waiting for {self.get_resource_type()} to be created.")
                waiter = self.get_service_client(aws_client).get_waiter(
                    "nodegroup_active"
                )
                waiter.wait(
                    clusterName=self.eks_cluster.name,
                    nodegroupName=self.name,
                    WaiterConfig={
                        "Delay": self.waiter_delay,
                        "MaxAttempts": self.waiter_max_attempts,
                    },
                )
            except Exception as e:
                print_error("Waiter failed.")
                print_error(e)
        return True

    def _read(self, aws_client: AwsApiClient) -> Optional[Any]:
        """Returns the EksNodeGroup

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        logger.debug(f"Reading {self.get_resource_type()}: {self.get_resource_name()}")

        from botocore.exceptions import ClientError

        service_client = self.get_service_client(aws_client)
        try:
            describe_response = service_client.describe_nodegroup(
                clusterName=self.eks_cluster.name,
                nodegroupName=self.name,
            )
            # logger.debug(f"describe_response: {describe_response}")
            nodegroup_dict = describe_response.get("nodegroup", {})

            self.created_at = nodegroup_dict.get("createdAt", None)
            self.nodegroup_status = nodegroup_dict.get("status", None)
            logger.debug(f"NodeGroup created_at: {self.created_at}")
            logger.debug(f"NodeGroup status: {self.nodegroup_status}")
            if self.created_at is not None:
                logger.debug(f"EksNodeGroup found: {self.name}")
                self.active_resource = describe_response
        except ClientError as ce:
            logger.debug(f"ClientError: {ce}")
        except Exception as e:
            print_error(f"Error reading {self.get_resource_type()}.")
            print_error(e)
        return self.active_resource

    def _delete(self, aws_client: AwsApiClient) -> bool:
        """Deletes the EksNodeGroup

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        print_info(f"Deleting {self.get_resource_type()}: {self.get_resource_name()}")

        # Step 1: Delete the IamRole
        if self.node_role_arn is None and self.create_role:
            nodegroup_iam_role = self.get_nodegroup_iam_role()
            try:
                nodegroup_iam_role.delete(aws_client)
            except Exception as e:
                print_error(
                    "IamRole deletion failed, please try again or delete manually"
                )
                print_error(e)

        # Step 2: Delete the NodeGroup
        service_client = self.get_service_client(aws_client)
        self.active_resource = None
        try:
            delete_response = service_client.delete_nodegroup(
                clusterName=self.eks_cluster.name,
                nodegroupName=self.name,
            )
            logger.debug(f"EksNodeGroup: {delete_response}")
            print_info(
                f"{self.get_resource_type()}: {self.get_resource_name()} deleted"
            )
            return True
        except Exception as e:
            print_error(f"{self.get_resource_type()} could not be deleted.")
            print_error("Please try again or delete resources manually.")
            print_error(e)
        return False

    def post_delete(self, aws_client: AwsApiClient) -> bool:

        # Wait for EksNodeGroup to be deleted
        if self.wait_for_deletion:
            try:
                print_info(f"Waiting for {self.get_resource_type()} to be deleted.")
                waiter = self.get_service_client(aws_client).get_waiter(
                    "nodegroup_deleted"
                )
                waiter.wait(
                    clusterName=self.eks_cluster.name,
                    nodegroupName=self.name,
                    WaiterConfig={
                        "Delay": self.waiter_delay,
                        "MaxAttempts": self.waiter_max_attempts,
                    },
                )
                return True
            except Exception as e:
                print_error("Waiter failed.")
                print_error(e)
        return True

    def get_nodegroup_iam_role(self) -> IamRole:
        """
        Create an IAM role and attach the required Amazon EKS IAM managed policy to it.
        """
        if self.node_role is not None:
            return self.node_role

        policy_arns = [
            "arn:aws:iam::aws:policy/AmazonEKSWorkerNodePolicy",
            "arn:aws:iam::aws:policy/AmazonEC2ContainerRegistryReadOnly",
            "arn:aws:iam::aws:policy/AmazonEKS_CNI_Policy",
            "arn:aws:iam::aws:policy/AmazonS3FullAccess",
            "arn:aws:iam::aws:policy/service-role/AmazonEBSCSIDriverPolicy",
        ]
        if self.add_policy_arns is not None and isinstance(self.add_policy_arns, list):
            policy_arns.extend(self.add_policy_arns)

        return IamRole(
            name=self.node_role_name or f"{self.name}-iam-role",
            assume_role_policy_document=dedent(
                """\
            {
              "Version": "2012-10-17",
              "Statement": [
                {
                  "Effect": "Allow",
                  "Principal": {
                    "Service": "ec2.amazonaws.com"
                  },
                  "Action": "sts:AssumeRole"
                }
              ]
            }
            """
            ),
            policy_arns=policy_arns,
        )



from phidata.app.superset.superset_webserver import (
    SupersetWebserver,
    ServiceType,
    ImagePullPolicy,
    RestartPolicy,
)
from phidata.app.superset.superset_init import SupersetInit
from phidata.app.superset.superset_worker import SupersetWorker
from phidata.app.superset.superset_worker_beat import SupersetWorkerBeat
from phidata.app.superset.superset_base import SupersetBase, SupersetBaseArgs
